/*
** Module   : AbstractQuery.java
** Abstract : Abstract base class for all query implementations
**
** Copyright (c) 2004-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20060102   @23793 Created initial version. Retrofitted existing
**                           query classes to extend this class. Provides
**                           basic accumulation architecture.
** 002 ECF 20060118   @23988 Modified to accommodate Accumulator changes.
**                           Accumulator break group categories are now
**                           managed internally by the Accumulator class.
** 003 ECF 20060221   @24714 Added validateSubstitutionArguments. Ensures
**                           all arguments are either BaseDataTypes or
**                           Resolvables.
** 004 ECF 20060404   @25365 Implement WrappedResource. Required for
**                           method and attribute support.
** 005 ECF 20060419   @25613 Added support for deferred mode accumulators.
**                           In this mode, the query signals to the
**                           accumulator that a loop iteration has
**                           occurred, and that the query has found a new
**                           value to process, but accumulation is
**                           deferred until there is an explicit trigger
**                           to process accumulation, typically invoked by
**                           business logic.
** 006 ECF 20060503   @25960 Added support for accumulation using external
**                           data sources. Enhanced integration of break
**                           group information.
** 007 ECF 20060504   @25972 Removed redundant interface. P2JQuery now
**                           extends WrappedResource, so it is no longer
**                           necessary to declare that we implement it
**                           here.
** 008 ECF 20060912   @29597 Minor debugging improvement. Added index of
**                           invalid argument to IllegalArgumentException
**                           message in validateSubstitutionArguments().
** 009 ECF 20060925   @29960 Added safety check to argument validation.
**                           If an argument is null and an error is
**                           pending with the ErrorManager, an exception
**                           is not raised. Added checkPendingError()
**                           method for subclasses.
** 010 ECF 20061003   @30150 Rewrote checkPendingError() to release
**                           buffers in the event of a pending error.
**                           Renamed method to prepareFetch().
** 011 ECF 20061023   @30605 Abstracted errorIfNull logic up from
**                           subclasses. Implement setErrorIfNull() from
**                           P2JQuery interface. Added isErrorIfNull()
**                           method for subclass use.
** 012 ECF 20061102   @30956 Implement new P2JQuery methods which support
**                           browse capability. Specifically, a query
**                           associated with a browse widget must fetch
**                           the next record automatically when it is
**                           repositioned. Also, a browse must be notified
**                           of "asynchronous" reposition events.
** 013 ECF 20061119   @31322 Additional reposition support. Added methods
**                           isFetchOnReposition() and
**                           hasRepositionListener().
** 014 ECF 20061204   @31572 Added browsed flag and is/setBrowsed methods.
**                           Indicates if query is associated with a UI
**                           browse widget. This is necessary because the
**                           query must behave differently when attached
**                           to a browse.
** 015 ECF 20061206   @31584 Added parameter to afterReposition(). Allows
**                           caller to override fetchOnReposition behavior
**                           to expressly disallow record fetch.
** 016 SIY 20070502   @33428 Added notifyRepositionListeners() method.
** 017 ECF 20070716   @34506 Implement new methods required by P2JQuery
**                           interface:
**                              isStandalone()
**                              setStandalone(boolean)
**                              cleanup()
** 018 ECF 20070820   @34904 Added support for temporary silent error
**                           mode. This allows query substitution parms to
**                           be processed safely during query construction
**                           and initialization.
** 019 SVL 20080418   @38037 Added the inverseSorting flag.
** 020 ECF 20080509   @38240 Added open() method. Necessary for converted
**                           OPEN QUERY statements.
** 021 ECF 20080509   @38250 Fixed regression in open(). Removed check for
**                           standalone flag.
** 022 SVL 20080514   @38276 reposition(1) if the query is browsed was
**                           removed from open().
** 023 ECF 20080515   @38287 Rolled back #022 (@38276). The reposition is
**                           necessary in cases where the query is
**                           assigned to a QueryWrapper after that wrapper
**                           was already registered to a browse widget. In
**                           that case, we cannot rely on the browse
**                           driving the reposition.
** 024 CA  20080624   @38891 Minor javadoc changes regarding standalone 
**                           field and accessors.
** 025 ECF 20080630   @39032 Minor memory optimization. Use Boolean
**                           constants instead of instantiating new
**                           Boolean objects.
** 026 ECF 20081106   @40319 Added default implementation of isOffEnd().
**                           Uses _isOffEnd() as its backing worker.
** 027 CA  20081210   @40835 Implemented CLOSE QUERY statement.
** 028 CA  20090227   @41390 Added WrappedResources.isUnknown method; this
**                           implementation returns false.
** 029 ECF 20090307   @42982 Fixed converted OPEN QUERY behavior. Added
**                           loadRowAtCursor() and getOffEnd(), changed
**                           afterReposition() to only fetch a record if the
**                           query is browsed.
** 030 ECF 20090707   @43103 Added isRepositionNotificationActive(). Reports
**                           whether we are currently notifying listeners of a
**                           reposition event.
** 031 SVL 20091030   @44291 getOffEnd() function was made public in order to
**                           support API change into P2JQuery.
** 032 SIY 20100303   @44701 Added hasAccumulators() helper method.
** 033 VMN 20130207          Added support for handle-based methods:
**                           GET-BUFFER-HANDLE (KW_GET_BUFH)
**                           GET-CURRENT (KW_GET_CUR),
**                           QUERY-CLOSE (KW_QRY_CLOS),
**                           REPOSITION-FORWARD (KW_REPOS_F),
**                           REPOSITION-TO-ROW (KW_REPOS_2R)
**                           REPOSITION-TO-ROWID (KW_REPOS_2I),
**                           and attributes
**                           NUM-RESULTS (KW_NUM_RES)
** 034 OM  20130212          Added implementation stubs for some dynamic methods.
** 035 VMN 20130213          Added support for attributes
**                           INDEX-INFORMATION (KW_IDX_INFO)
**                           FORWARD-ONLY (KW_FWD_ONLY)
**                           PREPARE_STRING (KW_PREP_STR)
** 036 VMN 20130219          Added support for attributes
**                           IS-OPEN (KW_IS_OPEN)
** 037 CA  20130221          Added support for UNIQUE-ID, ADM-DATA. Implemented Deletable. Added
**                           HandleChain support. 
** 038 OM  20130304          Refactored isValid and isUnknown of WrappedResource to valid 
**                           and unknown.
** 039 VMN 20130306          Added addBuffer(handle bufHandle) and corrected support
**                           for handle-based methods:
**                           GET-NEXT (KW_GET_NEXT)
**                           GET-PREV (KW_GET_PREV)
**                           GET-FIRST (KW_GET_1ST)
**                           GET-LAST (KW_GET_LAST)
**                           GET-CURRENT (KW_GET_CUR)
**                           Implemented
**                           getFirst(),getNext(),getLast(),getPrevious(),getCurrent()
** 040 CA  20130311          Added bufferHandle(integer).
** 041 VMN 20130311          Implemented queryClose, queryForward,
**                           queryReposition, queryRepositionByID,
**                           removed static methods
** 042 VMN 20130312          Corrected conversion and runtime for
**                           REPOSITION-TO-ROW (KW_REPOS_2R)
** 043 CA  20130315          Added missing INDEX-INFORMATION API.
** 044 OM  20130320          Added implementation for ResultListHandler interface.
** 045 VMN 20130321          Corrected conversion and runtime for
**                           NUM-BUFFERS attribute(KW_NUM_BUFF)
** 046 CA  20130322          Added missing queryForward(integer) API.
** 047 SVL 20130325          Added include and exclude functions.
** 048 CA  20130603          resourceDelete must return a boolean value.
** 049 CA  20130918          Added widget-pool support. The pool name must be validated before 
**                           anything else, and a null name means the default pool is used.
** 050 CA  20130927          Removed unknown() API, as by default the resource is known. 
** 051 ECF 20130405          Moved createQuery static methods to QueryWrapper. Cleaned up some
**                           return values.
** 052 CA  20131024          The legacy QUERY resource is implemented by QueryWrapper in P2J. 
**                           A block-level query is not compatible with the legacy QUERY resource,
**                           thus the implementation for almost all QUERY-related APIs is moved to
**                           QueryWrapper (what was not moved yet are the GET-FIRST/LAST/NEXT/PREV
**                           /CURRENT implementations).
** 053 ECF 20131028          Import change.
** 054 VMN 20140330          Added LockType errors processing.
*/

package com.goldencode.p2j.persist;

import java.util.*;
import org.apache.commons.logging.*;
import com.goldencode.p2j.persist.event.*;
import com.goldencode.p2j.persist.lock.*;
import com.goldencode.p2j.util.*;

/**
 * Abstract base class for all query implementation classes.  Implements the
 * <code>P2JQuery</code> interface and provides base services common to all
 * query implementations.
 */
public abstract class AbstractQuery
implements P2JQuery
{
   /** Logger */
   private static final Log LOG = Utils.getLog(AbstractQuery.class);
   
   /**
    * Indicates whether the query has the inverse sorting direction against
    * the original sorting. E.g.,
    * <pre>
    *  find next tt use-index si-name.
    *  find prev tt use-index si-name.
    * </pre>
    * will be converted to the code having a couple of AdaptiveFind queries:
    * <pre>
    *  AdaptiveFind query1 = new AdaptiveFind(tt, null, "tt.name asc");
    *  AdaptiveFind query2 = new AdaptiveFind(tt, null, "tt.name desc", true);
    * </pre>
    * and the second query will have the inverse sorting direction.
    */
   protected boolean inverseSorting = false;
   
   /** Flag indicating if this query was explicitly closed. */
   protected boolean closed = false;

   /** Should failed first/last/unique retrievals raise error? */
   private boolean errorIfNull = true;
   
   /** Does this query stand alone or is it contained by a query wrapper ? */
   private boolean standalone = true;
   
   /** Should the next row be fetched after a reposition? */
   private boolean fetchOnReposition = false;
   
   /** Is query associated with a UI browse widget? */
   private boolean browsed = false;
   
   /** Accumulators added to this query, mapped to their deferred status */
   private LinkedHashMap<Accumulator, Boolean> accumulators = null;
   
   /** Registry of listeners who are notified of query repositions */
   private WeakList<RepositionListener> repoRegistry = null;
   
   /** Flag to avoid infinite recursion in reposition notification */
   private boolean notificationActive = false;
   
   /**
    * Default constructor.
    */
   protected AbstractQuery()
   {
      init();
   }
   
   /**
    * Verify that an array of query substitution arguments contains only
    * <code>BaseDataType</code> and/or <code>Resolvable</code> elements.
    * <p>
    * A <code>null</code> argument is permissible only if the error manager
    * has an error pending.  We don't throw an exception in this case, as the
    * error will be processed later.
    *
    * @param   args
    *          Query substitution arguments to be validated.
    *
    * @throws  IllegalArgumentException
    *          if <code>args</code> is not <code>null</code>, but contains
    *          invalid argument types.
    */
   protected static void validateSubstitutionArguments(Object[] args)
   {
      if (args != null)
      {
         int len = args.length;
         for (int i = 0; i < len; i++)
         {
            Object next = args[i];
            
            // If there is an error pending, an argument may be null.  Allow
            // processing to continue.
            if (next == null && ErrorManager.isPendingError())
            {
               continue;
            }
            
            // Argument must be one of these types.
            if (!((next instanceof BaseDataType) ||
                  (next instanceof Resolvable)))
            {
               throw new IllegalArgumentException(
                  "Invalid where expression substitution argument:  " +
                  next +
                  " [" +
                  i +
                  "]");
            }
         }
      }
   }
   
   /**
    * Enable results list scrolling/repositioning.  This default
    * implementation does nothing.
    */
   public void setScrolling()
   {
   }
   
   /**
    * Reposition the prepared query to the first result if the query is being
    * browsed.
    */
   public void open()
   {
      if (browsed)
      {
         reposition(1);
      }
   }
   
   /**
    * Indicate whether the cursor has run off one or the other end of the
    * associated query's list of available results.  If <code>true</code>,
    * this indicates that the query cannot produce any more results in the
    * current scroll direction.
    *
    * @return  <code>true</code> if the cursor is off either end of its
    *          query's results list.
    */
   public logical isOffEnd()
   {
      boolean result = _isOffEnd();
      
      return (new logical(result));
   }
   
   /**
    * Add an accumulator to this query.  The accumulator expects to be
    * notified whenever a new result row is accessed (during sequential row
    * access).
    * <p>
    * This implementation assumes the accumulator is not associated with a
    * break group.
    *
    * @param   accum
    *          Accumulator to be registered with this query.
    */
   public void addAccumulator(Accumulator accum)
   {
      addAccumulator(accum, true);
   }
   
   /**
    * Add an accumulator to this query.  The accumulator expects to be
    * notified whenever a new result row is accessed (during sequential row
    * access).
    * <p>
    * This implementation assumes the accumulator is not associated with a
    * break group.
    *
    * @param   accum
    *          Accumulator to be registered with this query.
    * @param   deferred
    *          If <code>true</code>, calls to {@link #accumulate()} will not
    *          cause the accumulator to recalculate immediately, but instead
    *          to update the accumulator's state such that the next invocation
    *          of {@link com.goldencode.p2j.util.Accumulator#accumulate} will
    *          trigger the actual accumulation update.  If <code>false</code>,
    *          the accumulator recalculates its internal state immediately
    */
   public void addAccumulator(Accumulator accum, boolean deferred)
   {
      if (accumulators == null)
      {
         // Use sequential map to preserve add order.
         accumulators = new LinkedHashMap<Accumulator, Boolean>();
      }
      
      accumulators.put(accum, deferred ? Boolean.TRUE : Boolean.FALSE);
   }
   
   /**
    * Is this query set to fetch the next record after a reposition?
    *
    * @return  <code>true</code> if set to fetch after reposition, else
    *          <code>false</code>.
    */
   public boolean isFetchOnReposition()
   {
      return fetchOnReposition;
   }
   
   /**
    * Set this query to automatically fetch the next record after a
    * reposition, or turn this feature off.
    *
    * @param   enable
    *          <code>true</code> to enable automatic fetching;
    *          <code>false</code> to disable.
    */
   public void setFetchOnReposition(boolean enable)
   {
      fetchOnReposition = enable;
   }
   
   /**
    * Register a listener to receive notifications reposition events for this
    * query.
    *
    * @param   listener
    *          Reposition listener to be registered.
    */
   public void addRepositionListener(RepositionListener listener)
   {
      if (repoRegistry == null)
      { 
         repoRegistry = new WeakList<RepositionListener>();
      }
      
      repoRegistry.add(listener);
   }
   
   /**
    * Indicate to this query that whether it is associated with a browse
    * widget or not.
    *
    * @param   browsed
    *          <code>true</code> to indicate the query has an associated
    *          browse widget, else <code>false</code>.
    */
   public void setBrowsed(boolean browsed)
   {
      this.browsed = browsed;
   }

   /**
    * Notify all registered listeners that query is repositioned or closed.
    * 
    * @param   closed
    *          A <code>true</code> value will inform all listeners that the
    *          query was explicitly closed.
    */
   public void notifyRepositionListeners(boolean closed)
   {
      if (repoRegistry != null)
      {
         Iterator<RepositionListener> iter = repoRegistry.iterator();
         while (iter.hasNext())
         {
            RepositionListener next = iter.next();
            if (closed)
            {
               next.queryClosed();
            }
            else
            {
               next.queryRepositioned();
            }
         }
      }
   }
   
   /**
    * Indicate whether this query stands alone, or is wrapped by a query 
    * wrapper.    
    *  
    * @return  <code>true</code> if standalone, else <code>false</code>.
    */
   public boolean isStandalone()
   {
      return standalone;
   }
   
   /**
    * Change the standalone nature of this query.  A query is standalone by
    * default, but can be changed to non-standalone by the wrapper which 
    * contains it.
    * 
    * @param   standalone
    *          <code>true</code> to set query as standalone;
    *          <code>false</code> to set query as contained.
    */
   public void setStandalone(boolean standalone)
   {
      this.standalone = standalone;
   }

   /**
    * Clean up resources associated with this query when its scope ends.
    * <p>
    * This default implementation does nothing.
    */
   public void cleanup()
   {
   }
   
   /**
    * Explicitly close the prepared query.
    * <p>
    * This default implementation calls {@link #cleanup()}.
    */
   public void close()
   {
      cleanup();
      closed = true;
   }
   
   /**
    * Set a mode such that setting a null record into a backing buffer may or
    * may not result in an error condition.  This applies only to requests to
    * retrieve <code>FIRST</code>, <code>LAST</code>, and <code>UNIQUE</code>
    * records.
    *
    * @param   errorIfNull
    *          If <code>true</code>, setting a <code>null</code> value as the
    *          current record in an underlying buffer will raise an error
    *          condition (if not in silent error mode); if <code>false</code>,
    *          this action will raise an end condition instead.
    */
   public void setErrorIfNull(boolean errorIfNull)
   {
      this.errorIfNull = errorIfNull;
   }
   
   /**
    * Get the off-end status of this query, which indicates whether the query
    * has run off either end of its results.
    * 
    * @return  One of the {@link OffEnd} constants <code>NONE</code>,
    *          <code>FRONT</code>, or <code>BACK</code>.
    */
   public abstract OffEnd getOffEnd();
   
   /**
    * Indicate whether failed <code>FIRST</code>, <code>LAST</code>, and
    * <code>UNIQUE</code> retrieval attempts should be considered an error
    * condition.
    *
    * @return  <code>true</code> if failed first/last/unique retrievals
    *          should be considered an error;  <code>false</code> is such
    *          failures should be considered a query off-end condition.
    */
   protected boolean isErrorIfNull()
   {
      return errorIfNull;
   }
   
   /**
    * This method is invoked during construction to initialize the query.
    * Default behavior is to disable silent error mode, which may have been
    * enabled temporarily, just before the query's construction, in order
    * to suppress errors during query substitution parameter processing.
    * <p>
    * Query implementations which need to do this later (for instance, those
    * those which receive query substitution parameter values <i>after</i>
    * construction), may invoke this method at the appropriate time.  This
    * implementation is lenient with regard to being invoked multiple times;
    * Silent error mode will only be disabled once for each time it was
    * enabled.
    */
   protected void init()
   {
      RecordBuffer.disableSilentError();
   }
   
   /**
    * Report whether this query is associated with a browse widget.
    *
    * @return  <code>true</code> if associated with a browse, else
    *          <code>false</code>.
    */
   protected boolean isBrowsed()
   {
      return browsed;
   }
   
   /**
    * This method should be called by subclasses after any reposition of the
    * query.  If the query is set to fetch a record on reposition events, it
    * does so now (using the {@link #next} method).  If any listeners are
    * registered to receive notifications of reposition events, they are
    * notified now.
    *
    * @param   permitFetch
    *          <code>true</code> to permit immediate record fetch;
    *          <code>false</code> to override <code>fetchOnReposition</code>
    *          flag and explicitly disallow immediate record fetch.
    */
   protected void afterReposition(boolean permitFetch)
   {
      if (permitFetch && fetchOnReposition)
      {
         if (isBrowsed())
         {
            next();
         }
         else
         {
            loadRowAtCursor();
         }
      }
      
      // Prevent recursive notifications.
      if (isRepositionNotificationActive())
      {
         return;
      }
      
      notificationActive = true;
      notifyRepositionListeners(false);
      notificationActive = false;
   }
   
   /**
    * Indicate if we are currently notifying listeners of a reposition event.
    * 
    * @return  <code>true</code> if a reposition notification is active, else
    *          <code>false</code>.
    */
   protected boolean isRepositionNotificationActive()
   {
      return notificationActive;
   }
   
   /**
    * Synchronize the query with the cursor by causing the query to load the
    * current row's results (if any) into its backing buffer(s).
    * <p>
    * This implementation does nothing;  it should be overridden to implement
    * this behavior as needed.
    */
   protected void loadRowAtCursor()
   {
   }
   
   /**
    * Notify accumulators attached to this query (if any) that a new result
    * row has been accessed, and that they should update their internal state.
    * Subclasses should invoke this method whenever an iteration of a query
    * result occurs, such as on the retrieval of the first, last, next, or
    * previous result row.
    */
   protected final void accumulate()
   {
      if (!hasAccumulators())
         return;
      
      Iterator<Map.Entry<Accumulator, Boolean>> iter =
         accumulators.entrySet().iterator();
      while (iter.hasNext())
      {
         Map.Entry<Accumulator, Boolean> next = iter.next();
         Accumulator accum = next.getKey();
         Boolean deferred = next.getValue();
         
         accumulate(accum, deferred.booleanValue());
      }
   }
   
   /**
    * Check if query has attached accumulators.
    * 
    * @return  <code>true</code> if accumulators attached to the query.
    */
   protected boolean hasAccumulators()
   {
      return (accumulators == null || accumulators.isEmpty()) ? false : true;
   }
   
   /**
    * Subclasses which support the concept of break groups in query results
    * should override this method to report whether the current result row
    * being visited represents a new break group, for the given break group
    * key.
    *
    * @param   key
    *          Resolvable object which is used as the key to identify the
    *          break group category being tested.
    *
    * @return  <code>true</code> if the current result represents the first
    *          entry in a new break group;  else <code>false</code>.  This
    *          default implementation always returns <code>false</code>.
    */
   protected boolean isNewBreakGroup(Resolvable key)
   {
      return false;
   }
   
   /**
    * Determine whether the query result row currently being visited is the
    * last row within a break group whose category is identified by the
    * specified resolvable object.
    *
    * @param   key
    *          Break group category key.  May be <code>null</code> (which
    *          should always trigger a <code>false</code> return).
    *
    * @return  <code>false</code>;  subclasses should override this method if
    *          a different result is required.
    */
   protected boolean isLastOfBreakGroup(Resolvable key)
   {
      return false;
   }
   
   /**
    * Prepare to perform a fetch.  This involves checking for a pending error
    * in the error manager and releasing the current buffer record if an
    * error is pending.  Releasing the record may result in an exception
    * being thrown by the record buffer.
    *
    * @return  <code>false</code> if an error is pending, else
    *          <code>true</code>.
    */
   protected boolean prepareFetch()
   {
      if (ErrorManager.isPendingError())
      {
         try
         {
            releaseBuffers();
         }
         catch (QueryOffEndException exc)
         {
            // intentionally ignored
         }
         
         if (LOG.isDebugEnabled())
         {
            String msg = "Error pending in query";
            LOG.debug(msg, new ErrorConditionException(msg));
         }
         
         return false;
      }
      
      return true;
   }
   
   /**
    * Default, no-op implementation which should be overridden by subclasses
    * which need to release the current record of each of their buffers in
    * response to a fetch request while an error is pending in the error
    * manager.
    */
   protected void releaseBuffers()
   {
   }
   
   /**
    * Invoke the appropriate <code>iterate</code> method variant on the
    * specified accumulator.  If the accumulator manages only a primary
    * statistic (i.e., no break groups have been added), the {@link
    * com.goldencode.p2j.util.Accumulator#iterate(boolean) simple form}
    * is invoked.  If break groups have been added to the accumulator, the
    * {@link com.goldencode.p2j.util.Accumulator#iterate(java.util.Map,
    * boolean) more complex form} is invoked.
    *
    * @param   accum
    *          Target accumulator object.
    * @param   deferred
    *          <code>true</code> if the accumulator was registered to this
    *          query as a deferred accumulator, else <code>false</code>.
    *
    * @see     #addAccumulator
    */
   private void accumulate(Accumulator accum, boolean deferred)
   {
      Iterator<Resolvable> iter = accum.breakGroupKeys();
      
      // Iterate the accumulator.
      if (!iter.hasNext())
      {
         accum.iterate(deferred);
      }
      else
      {
         Map<Resolvable, Boolean> breakMap =
            new HashMap<Resolvable, Boolean>();
         while (iter.hasNext())
         {
            Resolvable key = iter.next();
            Boolean b = (isNewBreakGroup(key) ? Boolean.TRUE : Boolean.FALSE);
            breakMap.put(key, b);
         }
         accum.iterate(breakMap, deferred);
      }
      
      // Update the accumulator's "last break group" state.
      Resolvable key = accum.getFirstBreakGroupKey();
      accum.setLastOfGroup(isLastOfBreakGroup(key));
   }

   /**
    * Conversion of GET-BUFFER-HANDLE() method (KW_GET_BUFH).
    *
    * Returns a handle of a particular buffer of a query object.
    *
    * @param   bufferName
    *          buffer-name. A CHARACTER expression that evaluates to the name
    *          of a buffer in the query or ProDataSet object.
    *
    * @return buffer handle
    */
   public handle bufferHandle(String bufferName)
   {
      throw new RuntimeException("Must use the equivalent QueryWrapper API!");
   }

   /**
    * Conversion of GET-BUFFER-HANDLE() method (KW_GET_BUFH).
    *
    * Returns a handle of a particular buffer of a query object.
    *
    * @param   bufferName
    *          buffer-name. A CHARACTER expression that evaluates to the name
    *          of a buffer in the query or ProDataSet object.
    *
    * @return buffer handle
    */
   public handle bufferHandle(character bufferName)
   {
      throw new RuntimeException("Must use the equivalent QueryWrapper API!");
   }

   /**
    * Conversion of GET-BUFFER-HANDLE() method (KW_GET_BUFH).
    *
    * Returns a handle of a particular buffer of a query object.
    *
    * @param   bufferSequenceNumber
    *          buffer-sequence-number. An INTEGER that represents the sequence
    *          number of the desired buffer.
    *
    * @return buffer handle
    */
   public handle bufferHandle(int bufferSequenceNumber)
   {
      throw new RuntimeException("Must use the equivalent QueryWrapper API!");
   }

   /**
    * Conversion of GET-BUFFER-HANDLE() method (KW_GET_BUFH).
    *
    * Returns a handle of a particular buffer of a query object.
    *
    * @param   bufferSequenceNumber
    *          buffer-sequence-number. An INTEGER that represents the sequence
    *          number of the desired buffer.
    *
    * @return buffer handle
    */
   public handle bufferHandle(integer bufferSequenceNumber)
   {
      throw new RuntimeException("Must use the equivalent QueryWrapper API!");
   }

   /**
    * Conversion of QUERY-CLOSE() method (KW_QRY_CLOS).
    *
    * Closes a query object.
    *
    * @return  <code>true</code> if QUERY-CLOSE() succeeds,
    *          else <code>false</code>.
    */
   public logical queryClose()
   {
      throw new RuntimeException("Must use the equivalent QueryWrapper API!");
   }

   /**
    * Conversion of REPOSITION-FORWARD() method (KW_REPOS_F).
    *
    * Moves a query object's result list pointer forward a particular number
    * of rows.
    *
    * @return  <code>true</code> if REPOSITION-FORWARD() succeeds,
    *          else <code>false</code>.
    */
   public logical queryForward(integer rows)
   {
      throw new RuntimeException("Must use the equivalent QueryWrapper API!");
   }

   /**
    * Conversion of REPOSITION-FORWARD() method (KW_REPOS_F).
    *
    * Moves a query object's result list pointer forward a particular number
    * of rows.
    *
    * @return  <code>true</code> if REPOSITION-FORWARD() succeeds,
    *          else <code>false</code>.
    */
   public logical queryForward(int rows)
   {
      throw new RuntimeException("Must use the equivalent QueryWrapper API!");
   }

   /**
    * Conversion of REPOSITION-TO-ROW() method (KW_REPOS_2R).
    *
    * Moves a query object's result list pointer to the row corresponding to the specified
    * sequence number.
    *
    * @param   n
    *          An integer expression representing the sequence number.
    *
    * @return  <code>true</code> if REPOSITION-TO-ROW() succeeds,
    *          else <code>false</code>.
    */
   public logical queryReposition(int n)
   {
      throw new RuntimeException("Must use the equivalent QueryWrapper API!");
   }

   /**
    * Conversion of REPOSITION-TO-ROW() method (KW_REPOS_2R).
    *
    * Moves a query object's result list pointer to the row corresponding to the specified
    * sequence number.
    *
    * @param   n
    *          An integer expression representing the sequence number.
    *
    * @return  <code>true</code> if REPOSITION-TO-ROW() succeeds,
    *          else <code>false</code>.
    */
   public logical queryReposition(NumberType n)
   {
      throw new RuntimeException("Must use the equivalent QueryWrapper API!");
   }

   /**
    * Conversion of REPOSITION-TO-ROWID() method (KW_REPOS_2I).
    *
    * Reposition the cursor such that a request to retrieve the next result
    * will retrieve the result which matches the specified array of primary
    * key IDs.  A request to retrieve the previous result will retrieve the
    * result immediately previous to this, if any, in the results list.
    *
    * @param   id1
    *          A primary key ID, representing the target record (in the case
    *          of a join, this ID is associated with the left-most record in
    *          the join).
    * @param   joinIDs
    *          All remaining primary key IDs, if any, arranged from left to
    *          right to coincide with the records being joined by the
    *          underlying query.
    *
    * @return  <code>true</code> if REPOSITION-TO-ROWID() succeeds,
    *          else <code>false</code>.
    */
   public logical queryRepositionByID(rowid id1, rowid...joinIDs)
   {
      throw new RuntimeException("Must use the equivalent QueryWrapper API!");
   }

   /**
    * Conversion of NUM-RESULTS attribute (KW_NUM_RES).
    *
    * Getter of NUM-RESULTS attribute
    *
    */
   public integer getNumResults()
   {
      throw new RuntimeException("Must use the equivalent QueryWrapper API!");
   }

   /**
    * Adds one new buffer to a query object or dynamic ProDataSet object,
    * without affecting the other buffers, if any. Use the SET-BUFFERS( )
    * method to remove all prior buffers and set all buffers for the object
    * at the same time.
    *
    * @param   bufHandle
    *          A buffer handle to add.
    *
    * @return  true on success
    */
   public logical addBuffer(handle bufHandle)
   {
      throw new RuntimeException("Must use the equivalent QueryWrapper API!");
   }

   /**
    * Adds one new buffer to a query object or dynamic ProDataSet object, 
    * without affecting the other buffers, if any. Use the SET-BUFFERS( )
    * method to remove all prior buffers and set all buffers for the object 
    * at the same time.
    * 
    * @param   buf
    *          A buffer to add.
    *
    * @return  true on success
    */
   public logical addBuffer(Buffer buf)
   {
      throw new RuntimeException("Must use the equivalent QueryWrapper API!");
   }

   /**
    * Sets buffers for the query. Any buffers previously added
    * or set are removed. Use the <code>addBuffer()</code> method to add one 
    * buffer to the object, without affecting the other buffers, if any.
    * This method is the equivalent of <code>SET-BUFFERS()</code> method 
    * from 4GL.
    * 
    * @param   hBuffers
    *          Variable argument list of buffer handles to be set.
    *          In 4GL, the maximum number of buffers per query is 18.
    * 
    * @return  true on success
    */
   public logical setBuffers(handle ... hBuffers)
   {
      throw new RuntimeException("Must use the equivalent QueryWrapper API!");
   }

   /**
    * Sets buffers for the query. Any buffers previously added
    * or set are removed. Use the <code>addBuffer()</code> method to add one 
    * buffer to the object, without affecting the other buffers, if any.
    * This method is the equivalent of <code>SET-BUFFERS()</code> method 
    * from 4GL.
    * 
    * @param   bufs
    *          Variable argument list of Buffer to be set.
    *          In 4GL, the maximum number of buffers per query is 18.
    * 
    * @return  true on success
    */
   public logical setBuffers(Buffer ... bufs)
   {
      throw new RuntimeException("Must use the equivalent QueryWrapper API!");
   }

   /**
    * Compiles a predicate (query condition). The query must be open after this call.
    * This method is the equivalent of <code>QUERY-PREPARE()</code> method from 4GL.
    *  
    * @param   predicate
    *          The predicate to be prepared by this query. The syntax of the string is the same
    *          of the <code>OPEN QUERY</code> statement of the 4GL.
    * 
    * @return  true on success
    */
   public logical prepare(String predicate)
   {
      throw new RuntimeException("Must use the equivalent QueryWrapper API!");
   }

   /**
    * Compiles a predicate (query condition). The query must be open after this call.
    * This method is the equivalent of <code>QUERY-PREPARE()</code> method from 4GL.
    *  
    * @param   predicate
    *          The predicate to be prepared by this query. The syntax of the string is the same
    *          of the <code>OPEN QUERY</code> statement of the 4GL.
    * 
    * @return  true on success
    */
   public logical prepare(character predicate)
   {
      throw new RuntimeException("Must use the equivalent QueryWrapper API!");
   }

   /**
    * Opens a query object. 
    * This method is the equivalent of <code>QUERY-OPEN()</code> method from 4GL.
    * <u>Note</u>: You must perform <code>prepare()</code> on a query 
    * object before you perform <code>queryOpen<code> on it. 
    * 
    * @return  true on success
    */
   public logical queryOpen()
   {
      throw new RuntimeException("Must use the equivalent QueryWrapper API!");
   }

   /**
    * Conversion of INDEX-INFORMATION attribute (KW_IDX_INFO).
    *
    * Getter of INDEX-INFORMATION attribute
    *
    * @return A character string consisting of a comma-separated list of the index or
    *         indexes the query uses at the level of join specified.
    */
   public character indexInformation()
   {
      throw new RuntimeException("Must use the equivalent QueryWrapper API!");
   }

   /**
    * Conversion of INDEX-INFORMATION attribute (KW_IDX_INFO).
    *
    * Getter of INDEX-INFORMATION attribute
    *
    * @param n
    * An integer expression that evaluates to the level of join for which you want
    * index information.

    * @return A character string consisting of a comma-separated list of the index or
    *         indexes the query uses at the level of join specified.
    */
   public character indexInformation(int n)
   {
      throw new RuntimeException("Must use the equivalent QueryWrapper API!");
   }

   /**
    * Conversion of INDEX-INFORMATION attribute (KW_IDX_INFO).
    *
    * Getter of INDEX-INFORMATION attribute
    *
    * @param    n
    *           An integer expression that evaluates to the level of join for which you want index
    *           information.
    *
    * @return   A character string consisting of a comma-separated list of the index or
    *           indexes the query uses at the level of join specified.
    */
   public character indexInformation(NumberType n)
   {
      throw new RuntimeException("Must use the equivalent QueryWrapper API!");
   }

   /**
    * Conversion of FORWARD-ONLY attribute (KW_FWD_ONLY).
    *
    * Getter of FORWARD-ONLY attribute.
    *
    * @return current value of FORWARD-ONLY attribute.
    */
   public logical forwardOnly()
   {
      throw new RuntimeException("Must use the equivalent QueryWrapper API!");
   }

   /**
    * Conversion of FORWARD-ONLY attribute (KW_FWD_ONLY).
    *
    * Setter of FORWARD-ONLY attribute.
    *
    * @param  forwardOnly
    *         New value of FORWARD-ONLY attribute.
    */
   public void changeForwardOnly(logical forwardOnly)
   {
      throw new RuntimeException("Must use the equivalent QueryWrapper API!");
   }

   /**
    * Conversion of FORWARD-ONLY attribute (KW_FWD_ONLY).
    *
    * Setter of FORWARD-ONLY attribute.
    *
    * @param  forwardOnly
    *         New value of FORWARD-ONLY attribute.
    */
   public void changeForwardOnly(boolean forwardOnly)
   {
      throw new RuntimeException("Must use the equivalent QueryWrapper API!");
   }

   /**
    * Conversion of PREPARE_STRING attribute (KW_PREP_STR).
    *
    * Getter of PREPARE_STRING attribute.
    *
    * @return current value of PREPARE_STRING attribute.
    */
   public character prepareString()
   {
      throw new RuntimeException("Must use the equivalent QueryWrapper API!");
   }

   /**
    * Conversion of IS-OPEN attribute (KW_IS_OPEN).
    *
    * Getter of IS-OPEN attribute.
    *
    * @return current value of IS-OPEN attribute.
    */
   public logical isOpen()
   {
      throw new RuntimeException("Must use the equivalent QueryWrapper API!");
   }

   /**
    * Conversion of GET-NEXT method (KW_GET_NEXT).
    *
    * Moves a query object's result list pointer ahead one row.
    *
    * @return  <code>true</code> if the next record in the query is found. If the query is not
    *          open or the next record cannot be found (query is empty or the query result list
    *          pointer is on the last row), the method returns <code>false</code>. If the query
    *          object handle is invalid, the method returns the Unknown value (?).
    */
   public logical getNext()
   {
      boolean success = true;
      try
      {
         next();
      }
      catch (ErrorConditionException exc)
      {
         success = false;
      }
      return new logical(success);
   }

   /**
    * Conversion of GET-NEXT method (KW_GET_NEXT).
    *
    * Moves a query object's result list pointer ahead one row.
    *
    * @param   lockType
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if the next record in the query is found. If the query is not
    *          open or the next record cannot be found (query is empty or the query result list
    *          pointer is on the last row), the method returns <code>false</code>. If the query
    *          object handle is invalid, the method returns the Unknown value (?).
    */
   public logical getNext(LockType lockType)
   {
      return wrapNext(lockType);
   }

   /**
    * Conversion of GET-NEXT method (KW_GET_NEXT).
    *
    * Moves a query object's result list pointer ahead one row.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if the next record in the query is found. If the query is not
    *          open or the next record cannot be found (query is empty or the query result list
    *          pointer is on the last row), the method returns <code>false</code>. If the query
    *          object handle is invalid, the method returns the Unknown value (?).
    *
    * @throws  ErrorConditionException
    */
   public logical getNext(long lock)
   {
      LockType lockType = fromLockNowait(lock, 0);
      return lockType == null ? new logical(false) : wrapNext(lockType);
   }

   /**
    * Conversion of GET-NEXT method (KW_GET_NEXT).
    *
    * Moves a query object's result list pointer ahead one row.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if the next record in the query is found. If the query is not
    *          open or the next record cannot be found (query is empty or the query result list
    *          pointer is on the last row), the method returns <code>false</code>. If the query
    *          object handle is invalid, the method returns the Unknown value (?).
    *
    * @throws  ErrorConditionException
    */
   public logical getNext(NumberType lock)
   {
      LockType lockType = fromLockNowait(lock.longValue(), 0);
      return lockType == null ? new logical(false) : wrapNext(lockType);
   }

   /**
    * Conversion of GET-NEXT method (KW_GET_NEXT).
    *
    * Moves a query object's result list pointer ahead one row.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if the next record in the query is found. If the query is not
    *          open or the next record cannot be found (query is empty or the query result list
    *          pointer is on the last row), the method returns <code>false</code>. If the query
    *          object handle is invalid, the method returns the Unknown value (?).
    *
    * @throws  ErrorConditionException
    */
   public logical getNext(long lock, long nowait)
   {
      LockType lockType = fromLockNowait(lock, nowait);
      return lockType == null ? new logical(false) : wrapNext(lockType);
   }

   /**
    * Conversion of GET-NEXT method (KW_GET_NEXT).
    *
    * Moves a query object's result list pointer ahead one row.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if the next record in the query is found. If the query is not
    *          open or the next record cannot be found (query is empty or the query result list
    *          pointer is on the last row), the method returns <code>false</code>. If the query
    *          object handle is invalid, the method returns the Unknown value (?).
    *
    * @throws  ErrorConditionException
    */
   public logical getNext(long lock, NumberType nowait)
   {
      LockType lockType = fromLockNowait(lock, nowait.longValue());
      return lockType == null ? new logical(false) : wrapNext(lockType);
   }

   /**
    * Conversion of GET-NEXT method (KW_GET_NEXT).
    *
    * Moves a query object's result list pointer ahead one row.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if the next record in the query is found. If the query is not
    *          open or the next record cannot be found (query is empty or the query result list
    *          pointer is on the last row), the method returns <code>false</code>. If the query
    *          object handle is invalid, the method returns the Unknown value (?).
    *
    * @throws  ErrorConditionException
    */
   public logical getNext(NumberType lock, long nowait)
   {
      LockType lockType = fromLockNowait(lock.longValue(), nowait);
      return lockType == null ? new logical(false) : wrapNext(lockType);
   }

   /**
    * Conversion of GET-NEXT method (KW_GET_NEXT).
    *
    * Moves a query object's result list pointer ahead one row.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if the next record in the query is found. If the query is not
    *          open or the next record cannot be found (query is empty or the query result list
    *          pointer is on the last row), the method returns <code>false</code>. If the query
    *          object handle is invalid, the method returns the Unknown value (?).
    *
    * @throws  ErrorConditionException
    */
   public logical getNext(NumberType lock, NumberType nowait)
   {
      LockType lockType = fromLockNowait(lock.longValue(), nowait.longValue());
      return lockType == null ? new logical(false) : wrapNext(lockType);
   }

   /**
    * Conversion of GET-PREV method (KW_GET_PREV).
    *
    * Moves a query object's result list pointer back one row.
    *
    * @return  <code>true</code> if the previous record in the query is found. If the query is
    *          not open or the previous record cannot be found (query is empty or the query
    *          result list pointer is on the first row), the method returns <code>false</code>.
    *          If the query object handle is invalid, the method returns the Unknown value (?).
    */
   public logical getPrevious()
   {
      boolean success = true;
      try
      {
         previous();
      }
      catch (ErrorConditionException exc)
      {
         success = false;
      }
      return new logical(success);
   }

   /**
    * Conversion of GET-PREV method (KW_GET_PREV).
    *
    * Moves a query object's result list pointer back one row.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if the previous record in the query is found. If the query is
    *          not open or the previous record cannot be found (query is empty or the query
    *          result list pointer is on the first row), the method returns <code>false</code>.
    *          If the query object handle is invalid, the method returns the Unknown value (?).
    *
    * @throws  ErrorConditionException
    */
   public logical getPrevious(long lock)
   {
      LockType lockType = fromLockNowait(lock, 0);
      return lockType == null ? new logical(false) : wrapPrevious(lockType);
   }

   /**
    * Conversion of GET-PREV method (KW_GET_PREV).
    *
    * Moves a query object's result list pointer back one row.
    *
    * @param   lockType
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if the previous record in the query is found. If the query is
    *          not open or the previous record cannot be found (query is empty or the query
    *          result list pointer is on the first row), the method returns <code>false</code>.
    *          If the query object handle is invalid, the method returns the Unknown value (?).
    *
    * @throws  IllegalArgumentException
    *          if <code>lock</code> has not valid internal value.
    */
   public logical getPrevious(LockType lockType)
   {
      return wrapPrevious(lockType);
   }

   /**
    * Conversion of GET-PREV method (KW_GET_PREV).
    *
    * Moves a query object's result list pointer back one row.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if the previous record in the query is found. If the query is
    *          not open or the previous record cannot be found (query is empty or the query
    *          result list pointer is on the first row), the method returns <code>false</code>.
    *          If the query object handle is invalid, the method returns the Unknown value (?).
    *
    * @throws  ErrorConditionException
    */
   public logical getPrevious(NumberType lock)
   {
      LockType lockType = fromLockNowait(lock.longValue(), 0);
      return lockType == null ? new logical(false) : wrapPrevious(lockType);
   }

   /**
    * Conversion of GET-PREV method (KW_GET_PREV).
    *
    * Moves a query object's result list pointer back one row.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if the previous record in the query is found. If the query is
    *          not open or the previous record cannot be found (query is empty or the query
    *          result list pointer is on the first row), the method returns <code>false</code>.
    *          If the query object handle is invalid, the method returns the Unknown value (?).
    *
    * @throws  ErrorConditionException
    */
   public logical getPrevious(long lock, long nowait)
   {
      LockType lockType = fromLockNowait(lock, nowait);
      return lockType == null ? new logical(false) : wrapPrevious(lockType);
   }

   /**
    * Conversion of GET-PREV method (KW_GET_PREV).
    *
    * Moves a query object's result list pointer back one row.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if the previous record in the query is found. If the query is
    *          not open or the previous record cannot be found (query is empty or the query
    *          result list pointer is on the first row), the method returns <code>false</code>.
    *          If the query object handle is invalid, the method returns the Unknown value (?).
    *
    * @throws  ErrorConditionException
    */
   public logical getPrevious(NumberType lock, long nowait)
   {
      LockType lockType = fromLockNowait(lock.longValue(), nowait);
      return lockType == null ? new logical(false) : wrapPrevious(lockType);
   }

   /**
    * Conversion of GET-PREV method (KW_GET_PREV).
    *
    * Moves a query object's result list pointer back one row.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if the previous record in the query is found. If the query is
    *          not open or the previous record cannot be found (query is empty or the query
    *          result list pointer is on the first row), the method returns <code>false</code>.
    *          If the query object handle is invalid, the method returns the Unknown value (?).
    *
    * @throws  ErrorConditionException
    */
   public logical getPrevious(long lock, NumberType nowait)
   {
      LockType lockType = fromLockNowait(lock, nowait.longValue());
      return lockType == null ? new logical(false) : wrapPrevious(lockType);
   }

   /**
    * Conversion of GET-PREV method (KW_GET_PREV).
    *
    * Moves a query object's result list pointer back one row.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if the previous record in the query is found. If the query is
    *          not open or the previous record cannot be found (query is empty or the query
    *          result list pointer is on the first row), the method returns <code>false</code>.
    *          If the query object handle is invalid, the method returns the Unknown value (?).
    *
    * @throws  ErrorConditionException
    */
   public logical getPrevious(NumberType lock, NumberType nowait)
   {
      LockType lockType = fromLockNowait(lock.longValue(), nowait.longValue());
      return lockType == null ? new logical(false) : wrapPrevious(lockType);
   }

   /**
    * Conversion of GET-FIRST method (KW_GET_1ST).
    *
    * Moves a query object's result list pointer to the first row.
    *
    * @return  <code>true</code> if the first record in the query is found. If the query is not
    *          open or the first record cannot be found (query is empty), the method returns
    *          <code>false</code>. If the query object handle is invalid, the method returns
    *          the Unknown value (?).
    */
   public logical getFirst()
   {
      boolean success = true;
      try
      {
         first();
      }
      catch (ErrorConditionException exc)
      {
         success = false;
      }
      return new logical(success);
   }

   /**
    * Conversion of GET-FIRST method (KW_GET_1ST).
    *
    * Moves a query object's result list pointer to the first row.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if the first record in the query is found. If the query is not
    *          open or the first record cannot be found (query is empty), the method returns
    *          <code>false</code>. If the query object handle is invalid, the method returns
    *          the Unknown value (?).
    *
    * @throws  IllegalArgumentException
    *          if <code>lock</code> has not valid internal value.
    *
    * @throws  ErrorConditionException
    */
   public logical getFirst(long lock)
   {
      LockType lockType = fromLockNowait(lock, 0);
      return lockType == null ? new logical(false) : wrapFirst(lockType);
   }

   /**
    * Conversion of GET-FIRST method (KW_GET_1ST).
    *
    * Moves a query object's result list pointer to the first row.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if the first record in the query is found. If the query is not
    *          open or the first record cannot be found (query is empty), the method returns
    *          <code>false</code>. If the query object handle is invalid, the method returns
    *          the Unknown value (?).
    *
    * @throws  ErrorConditionException
    */
   public logical getFirst(NumberType lock)
   {
      LockType lockType = fromLockNowait(lock.longValue(), 0);
      return lockType == null ? new logical(false) : wrapFirst(lockType);
   }

   /**
    * Conversion of GET-FIRST method (KW_GET_1ST).
    *
    * Moves a query object's result list pointer to the first row.
    *
    * @param   lockType
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if the first record in the query is found. If the query is not
    *          open or the first record cannot be found (query is empty), the method returns
    *          <code>false</code>. If the query object handle is invalid, the method returns
    *          the Unknown value (?).
    *
    * @throws  IllegalArgumentException
    *          if <code>lock</code> has not valid internal value.
    */
   public logical getFirst(LockType lockType)
   {
      return wrapFirst(lockType);
   }

   /**
    * Conversion of GET-FIRST method (KW_GET_1ST).
    *
    * Moves a query object's result list pointer to the first row.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if the first record in the query is found. If the query is not
    *          open or the first record cannot be found (query is empty), the method returns
    *          <code>false</code>. If the query object handle is invalid, the method returns
    *          the Unknown value (?).
    *
    * @throws  IllegalArgumentException
    *          if <code>lock</code> or <code>nowait</code> have not valid internal values.
    *
    * @throws  ErrorConditionException
    */
   public logical getFirst(long lock, long nowait)
   {
      LockType lockType = fromLockNowait(lock, nowait);
      return lockType == null ? new logical(false) : wrapFirst(lockType);
   }

   /**
    * Conversion of GET-FIRST method (KW_GET_1ST).
    *
    * Moves a query object's result list pointer to the first row.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if the first record in the query is found. If the query is not
    *          open or the first record cannot be found (query is empty), the method returns
    *          <code>false</code>. If the query object handle is invalid, the method returns
    *          the Unknown value (?).
    *
    * @throws  ErrorConditionException
    */
   public logical getFirst(NumberType lock, long nowait)
   {
      LockType lockType = fromLockNowait(lock.longValue(), nowait);
      return lockType == null ? new logical(false) : wrapFirst(lockType);
   }

   /**
    * Conversion of GET-FIRST method (KW_GET_1ST).
    *
    * Moves a query object's result list pointer to the first row.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if the first record in the query is found. If the query is not
    *          open or the first record cannot be found (query is empty), the method returns
    *          <code>false</code>. If the query object handle is invalid, the method returns
    *          the Unknown value (?).
    *
    * @throws  IllegalArgumentException
    *          if <code>lock</code> or <code>nowait</code> have not valid internal values.
    *
    * @throws  ErrorConditionException
    */
   public logical getFirst(long lock, NumberType nowait)
   {
      LockType lockType = fromLockNowait(lock, nowait.longValue());
      return lockType == null ? new logical(false) : wrapFirst(lockType);
   }

   /**
    * Conversion of GET-FIRST method (KW_GET_1ST).
    *
    * Moves a query object's result list pointer to the first row.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if the first record in the query is found. If the query is not
    *          open or the first record cannot be found (query is empty), the method returns
    *          <code>false</code>. If the query object handle is invalid, the method returns
    *          the Unknown value (?).
    *
    * @throws  ErrorConditionException
    */
   public logical getFirst(NumberType lock, NumberType nowait)
   {
      LockType lockType = fromLockNowait(lock.longValue(), nowait.longValue());
      return lockType == null ? new logical(false) : wrapFirst(lockType);
   }

   /**
    * Conversion of GET-LAST method (KW_GET_LAST).
    *
    * Moves a query object's result list pointer to the last row.
    *
    * @return  <code>true</code> if the last record in the query is found. If the query is not
    *          open or the last record cannot be found (query is empty), the method returns
    *          <code>false</code>. If the query object handle is invalid, the method returns
    *          the Unknown value (?).
    */
   public logical getLast()
   {
      boolean success = true;
      try
      {
         last();
      }
      catch (ErrorConditionException exc)
      {
         success = false;
      }
      return new logical(success);
   }

   /**
    * Conversion of GET-LAST method (KW_GET_LAST).
    *
    * Moves a query object's result list pointer to the last row.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if the last record in the query is found. If the query is not
    *          open or the last record cannot be found (query is empty), the method returns
    *          <code>false</code>. If the query object handle is invalid, the method returns
    *          the Unknown value (?).
    *
    * @throws  ErrorConditionException
    */
   public logical getLast(long lock)
   {
      LockType lockType = fromLockNowait(lock, 0);
      return lockType == null ? new logical(false) : wrapLast(lockType);
   }

   /**
    * Conversion of GET-LAST method (KW_GET_LAST).
    *
    * Moves a query object's result list pointer to the last row.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if the last record in the query is found. If the query is not
    *          open or the last record cannot be found (query is empty), the method returns
    *          <code>false</code>. If the query object handle is invalid, the method returns
    *          the Unknown value (?).
    *
    * @throws  ErrorConditionException
    */
   public logical getLast(NumberType lock)
   {
      LockType lockType = fromLockNowait(lock.longValue(), 0);
      return lockType == null ? new logical(false) : wrapLast(lockType);
   }

   /**
    * Conversion of GET-LAST method (KW_GET_LAST).
    *
    * Moves a query object's result list pointer to the last row.
    *
    * @param   lockType
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if the last record in the query is found. If the query is not
    *          open or the last record cannot be found (query is empty), the method returns
    *          <code>false</code>. If the query object handle is invalid, the method returns
    *          the Unknown value (?).
    *
    * @throws  IllegalArgumentException
    *          if <code>lock</code> has not valid internal value.
    */
   public logical getLast(LockType lockType)
   {
      return wrapLast(lockType);
   }

   /**
    * Conversion of GET-LAST method (KW_GET_LAST).
    *
    * Moves a query object's result list pointer to the last row.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if the last record in the query is found. If the query is not
    *          open or the last record cannot be found (query is empty), the method returns
    *          <code>false</code>. If the query object handle is invalid, the method returns
    *          the Unknown value (?).
    *
    * @throws  ErrorConditionException
    */
   public logical getLast(long lock, long nowait)
   {
      LockType lockType = fromLockNowait(lock, nowait);
      return lockType == null ? new logical(false) : wrapLast(lockType);
   }

   /**
    * Conversion of GET-LAST method (KW_GET_LAST).
    *
    * Moves a query object's result list pointer to the last row.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if the last record in the query is found. If the query is not
    *          open or the last record cannot be found (query is empty), the method returns
    *          <code>false</code>. If the query object handle is invalid, the method returns
    *          the Unknown value (?).
    *
    * @throws  ErrorConditionException
    */
   public logical getLast(NumberType lock, long nowait)
   {
      LockType lockType = fromLockNowait(lock.longValue(), nowait);
      return lockType == null ? new logical(false) : wrapLast(lockType);
   }

   /**
    * Conversion of GET-LAST method (KW_GET_LAST).
    *
    * Moves a query object's result list pointer to the last row.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if the last record in the query is found. If the query is not
    *          open or the last record cannot be found (query is empty), the method returns
    *          <code>false</code>. If the query object handle is invalid, the method returns
    *          the Unknown value (?).
    *
    * @throws  ErrorConditionException
    */
   public logical getLast(long lock, NumberType nowait)
   {
      LockType lockType = fromLockNowait(lock, nowait.longValue());
      return lockType == null ? new logical(false) : wrapLast(lockType);
   }

   /**
    * Conversion of GET-LAST method (KW_GET_LAST).
    *
    * Moves a query object's result list pointer to the last row.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if the last record in the query is found. If the query is not
    *          open or the last record cannot be found (query is empty), the method returns
    *          <code>false</code>. If the query object handle is invalid, the method returns
    *          the Unknown value (?).
    *
    * @throws  ErrorConditionException
    */
   public logical getLast(NumberType lock, NumberType nowait)
   {
      LockType lockType = fromLockNowait(lock.longValue(), nowait.longValue());
      return lockType == null ? new logical(false) : wrapLast(lockType);
   }

   /**
    * Conversion of GET-CURRENT() method (KW_GET_CUR).
    *
    * Refetches the current record or records associated with the query.
    *
    * @return  <code>true</code> if GET-CURRENT() succeeds, else <code>false</code>.
    */
   public logical getCurrent()
   {
      boolean success = true;
      try
      {
         current();
      }
      catch (ErrorConditionException exc)
      {
         success = false;
      }
      return new logical(success);
   }

   /**
    * Conversion of GET-CURRENT() method (KW_GET_CUR).
    *
    * Refetches the current record or records associated with the query.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if GET-CURRENT() succeeds, else <code>false</code>.
    *
    * @throws  ErrorConditionException
    */
   public logical getCurrent(long lock)
   {
      LockType lockType = fromLockNowait(lock, 0);
      return lockType == null ? new logical(false) : wrapCurrent(lockType);
   }

   /**
    * Conversion of GET-CURRENT() method (KW_GET_CUR).
    *
    * Refetches the current record or records associated with the query.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if GET-CURRENT() succeeds, else <code>false</code>.
    *
    * @throws  ErrorConditionException
    */
   public logical getCurrent(NumberType lock)
   {
      LockType lockType = fromLockNowait(lock.longValue(), 0);
      return lockType == null ? new logical(false) : wrapCurrent(lockType);
   }

   /**
    * Conversion of GET-CURRENT() method (KW_GET_CUR).
    *
    * Refetches the current record or records associated with the query.
    *
    * @param   lockType of LockType
    *
    * @return  <code>true</code> if GET-CURRENT() succeeds, else <code>false</code>.
    */
   public logical getCurrent(LockType lockType)
   {
      return wrapCurrent(lockType);
   }

   /**
    * Conversion of GET-CURRENT() method (KW_GET_CUR).
    *
    * Refetches the current record or records associated with the query.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if GET-CURRENT() succeeds, else <code>false</code>.
    *
    * @throws  ErrorConditionException
    */
   public logical getCurrent(long lock, long nowait)
   {
      LockType lockType = fromLockNowait(lock, nowait);
      return lockType == null ? new logical(false) : wrapCurrent(lockType);
   }

   /**
    * Conversion of GET-CURRENT() method (KW_GET_CUR).
    *
    * Refetches the current record or records associated with the query.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if GET-CURRENT() succeeds, else <code>false</code>.
    *
    * @throws  ErrorConditionException
    */
   public logical getCurrent(NumberType lock, long nowait)
   {
      LockType lockType = fromLockNowait(lock.longValue(), nowait);
      return lockType == null ? new logical(false) : wrapCurrent(lockType);
   }

   /**
    * Conversion of GET-CURRENT() method (KW_GET_CUR).
    *
    * Refetches the current record or records associated with the query.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if GET-CURRENT() succeeds, else <code>false</code>.
    *
    * @throws  ErrorConditionException
    */
   public logical getCurrent(long lock, NumberType nowait)
   {
      LockType lockType = fromLockNowait(lock, nowait.longValue());
      return lockType == null ? new logical(false) : wrapCurrent(lockType);
   }

   /**
    * Conversion of GET-CURRENT() method (KW_GET_CUR).
    *
    * Refetches the current record or records associated with the query.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if GET-CURRENT() succeeds, else <code>false</code>.
    *
    * @throws  ErrorConditionException
    */
   public logical getCurrent(NumberType lock, NumberType nowait)
   {
      LockType lockType = fromLockNowait(lock.longValue(), nowait.longValue());
      return lockType == null ? new logical(false) : wrapCurrent(lockType);
   }

   /**
    * Specifies the fields included in a record retrieval.
    *
    * @param dmo
    *        DMO proxy which defines the buffer for which the fields are specified.
    * @param fields
    *        The fields included in a record retrieval. If no fields are specified, only record
    *        identifiers are retrieved. No subscripts can be specified for extent fields.
    */
   public void include(DataModelObject dmo, String ... fields)
   {
      // TODO: implement
   }

   /**
    * Specifies the fields excluded from a record retrieval.
    *
    * @param dmo
    *        DMO proxy which defines the buffer for which the fields are specified.
    * @param fields
    *        The fields excluded from a record retrieval. If no fields are specified, complete
    *        records are retrieved. No subscripts can be specified for extent fields.
    */
   public void exclude(DataModelObject dmo, String ... fields)
   {
      // TODO: implement
   }

   /**
    * Conversion of NUM-BUFFERS() method (KW_NUM_BUFF).
    *
    * @return  the number of buffers in a query or ProDataSet object.
    */
   public integer numBuffers()
   {
      return new integer(getTableCount());
   }

   /**
    * Return LockType instance for the given internal lock and nowait with corresponding query
    * processing .
    * <p>
    *
    * @param   lock
    *          Integer value.
    * @param   nowait
    *          Integer value.
    *
    * @return  <code>LockType</code> instance.
    *
    * @throws   ErrorConditionException
    */
   private static LockType fromLockNowait(long lock, long nowait)
   throws ErrorConditionException
   {
      LockType lockType = null;
      try
      {
         lockType = LockType.fromLockNowait(lock, nowait);
      }
      catch (LockTypeException e)
      {
         String err;
         int num;
         if (e.isNowaitCorrect())
         {
            err = "The first argument to QUERY GET methods must be " +
                  "NO-LOCK, SHARE-LOCK or EXCLUSIVE-LOCK";
            num = 7314;
         }
         else
         {
            err = "QUERY GET second argument must be NO-WAIT or 0";
            num = 7360;
         }
         ErrorManager.recordOrThrowError(num, err);
      }
      return lockType;
   }

   /**
    * Worker method for wrapping ErrorConditionException and reusing
    * in all P2JQuery.getNext() implementations
    *
    * @param   lockType
    *          instance of LockType
    *
    * @return  <code>true</code> if GET-NEXT() succeeds, else <code>false</code>.
    */
   private logical wrapNext(LockType lockType)
   {
      boolean success = true;
      try
      {
         next(lockType);
      }
      catch (ErrorConditionException exc)
      {
         success = false;
      }
      return new logical(success);
   }

   /**
    * Worker method for wrapping ErrorConditionException and reusing
    * in all P2JQuery.getPrevious() implementations
    *
    * @param   lockType
    *          instance of LockType
    *
    * @return  <code>true</code> if GET-PREV() succeeds, else <code>false</code>.
    */
   private logical wrapPrevious(LockType lockType)
   {
      boolean success = true;
      try
      {
         previous(lockType);
      }
      catch (ErrorConditionException exc)
      {
         success = false;
      }
      return new logical(success);
   }

   /**
    * Worker method for wrapping ErrorConditionException and reusing
    * in all P2JQuery.getFirst() implementations
    *
    * @param   lockType
    *          instance of LockType
    *
    * @return  <code>true</code> if GET-FIRST() succeeds, else <code>false</code>.
    */
   private logical wrapFirst(LockType lockType)
   {
      boolean success = true;
      try
      {
         first(lockType);
      }
      catch (ErrorConditionException exc)
      {
         success = false;
      }
      return new logical(success);
   }

   /**
    * Worker method for wrapping ErrorConditionException and reusing
    * in all P2JQuery.getLast() implementations
    *
    * @param   lockType
    *          instance of LockType
    *
    * @return  <code>true</code> if GET-LAST() succeeds, else <code>false</code>.
    */
   private logical wrapLast(LockType lockType)
   {
      boolean success = true;
      try
      {
         last(lockType);
      }
      catch (ErrorConditionException exc)
      {
         success = false;
      }
      return new logical(success);
   }

   /**
    * Worker method for wrapping ErrorConditionException and reusing
    * in all P2JQuery.getCurrent() implementations
    *
    * @param   lockType
    *          instance of LockType
    *
    * @return  <code>true</code> if GET-CURRENT() succeeds, else <code>false</code>.
    */
   private logical wrapCurrent(LockType lockType)
   {
      boolean success = true;
      try
      {
         current(lockType);
      }
      catch (ErrorConditionException exc)
      {
         success = false;
      }
      return new logical(success);
   }
}
