/*
** Module   : BufferImpl.java
** Abstract : implementation of Buffer interface
**
** Copyright (c) 2013-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description----------------------------------
** 001 ECF 20130118 Created initial version. Implementation of the Buffer
**                  interface; provides services for buffer-related, Progress
**                  methods, attributes, and some language statements.
** 002 VMN 20130207 Added support for handle-based methods
**                  BUFFER-COPY (KW_BUF_COPY)
**                  BUFFER-COMPARE (KW_BUF_COMP)
**                  BUFFER-FIELD (KW_BUF_FLD)
**                  BUFFER-RELEASE (KW_BUF_REL)
**                  FIND-BY-ROWID (KW_FIND_BR)
**                  FIND-UNIQUE (KW_GET_UNI)
**                  FIND-FIRST (KW_FIND_1ST)
**                  FIND-LAST (KW_FIND_LST)
**                  and attributes
**                  NUM-FIELDS (KW_NUM_FLD)
**                  COLUMN-LABEL (KW_COL_LAB)
**                  INITIAL (KW_INIT)
**                  AMBIGUOUS (KW_AMBIG)
** 003 ECF 20130214 Added bufferCreate and variants of bufferCopy.
** 004 SVL 20130213 Added disable*Triggers functions.
** 005 VMN 20130217 Added support for handle-based attributes:
**                  CURRENT-ITERATION (KW_CUR_ITER)
**                  ERROR (KW_ERROR)
**                  DBNAME (KW_DBNAME)
**                  TABLE (KW_TABLE)
** 006 VMN 20130219 Added support for attributes
**                  TABLE-HANDLE (KW_TAB_HAND)
** 007 CA  20130221 Added support for ADM-DATA, UNIQUE-ID, NAMESPACE-URI. Implemented Deletable.
**                  Added missing BUFFER-FIELD stubs.
** 008 GES 20130224 Added support for a DELETE statement with a validation expression.
** 009 ECF 20130224 Added limited support for embedded SQL.
** 010 OM  20130222 Added bufferField variants and dereference from Dereferenceable.
** 011 OM  20130304 Refactored isValid and isUnknown of WrappedResource to valid and unknown.
** 012 VMN 20130306 Corrected support for handle-based methods:
**                  FIND-UNIQUE (KW_GET_UNI)
**                  FIND-FIRST (KW_FIND_1ST)
**                  FIND-LAST (KW_FIND_LST)
** 013 VMN 20130311 Corrected support and runtime for handle-based methods:
**                  FIND-UNIQUE (KW_GET_UNI)
**                  FIND-FIRST (KW_FIND_1ST)
**                  FIND-LAST (KW_FIND_LST)
**                  FIND-BY-ROWID (KW_FIND_BR)
**                  BUFFER-RELEASE (KW_BUF_REL)
**                  BUFFER-COMPARE (KW_BUF_COMP)
** 014 ECF 20130309 Changed API for bufferCopy and bufferCompare; only wrapped data types are
**                  allowed for literal parameters, to prevent the explosion of the API. 
** 015 VMN 20130314 Implemented runtime for handle-based methods:
**                  BUFFER-COMPARE (KW_BUF_COMP) also noLobs parameter type changed
**                  BUFFER-COPY (KW_BUF_COPY) (simple signatures)
**                  First parameter type changed from String to character for
**                  FIND-UNIQUE (KW_GET_UNI)
**                  FIND-FIRST (KW_FIND_1ST)
**                  FIND-LAST (KW_FIND_LST)
** 016 CA  20130323 Changed the APIs for KW_BUF_DEL and KW_EMPTY_TT to return logical.
** 017 ECF 20130417 Moved initial() method to FieldReference.
** 018 VMN 20130426 Implemented runtime for handle-based methods:
**                  BUFFER-COMPARE (KW_BUF_COMP)
**                  BUFFER-COPY (KW_BUF_COPY)
** 019 CA  20130603 resourceDelete must return a boolean value.
** 020 VMN 20130711 Implemented verifyTransaction() and added to handle-based methods:
**                  BUFFER-CREATE (KW_BUF_CREA)
**                  BUFFER-DELETE (KW_BUF_DELETE)
**                  FIND-BY-ROWID (KW_FIND_BR)
** 021 VMN 20130717 Added aggregated functions support for embedded SQL.
** 022 SVL 20130822 Added DYNAMIC attribute, implemented bufferField.
** 023 CA  20130927 Removed unknown() API, as by default the resource is known. 
** 024 CA  20131021 Do not delete static buffers.
** 025 ECF 20131028 Import change.
** 026 SVL 20131028 Handled bufferField() for static tables, added doSetName.
** 027 CA  20131013 Static buffers are deleted only when the instantiating procedure gets deleted.
** 028 OM  20131128 Stub implementation for getUniqueID.
** 029 HC  20131222 Implemented ADM-DATA and UNIQUE-ID attribute.
** 030 VMN 20140110 Implemented bufferField(character).
** 031 OM  20140107 Implemented disableDumpTriggers() and disableLoadTriggers() methods and 
**                  and removed unneded openReadOnlyScope().
** 032 OM  20140116 Implemeneted attributes and methods (NUM-FIELDS, ERROR, DBNAME, TABLE,
**                  NAMESPACE-URI, referefence(::), BUFFER-FIELD(int), intermediary step for
**                  findFirst, findLast, findUnique).
** 033 OM  20140122 Added primary index selection for find-by-rowid dynamic finds. Added 
**                  implementation for dynamic finds methods using DynamicQueryHelper.
** 034 ECF 20140305 Added defaultBufferHandle implementation, reworked bufferField(String) to
**                  handle invalid buffer name.
** 035 VMN 20140330 Added LockType errors processing.
*/

package com.goldencode.p2j.persist;

import com.goldencode.p2j.persist.lock.*;
import com.goldencode.p2j.persist.trigger.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.util.UniqueIdGenerator.IdKind;

/**
 * This class implements services representing Progress builtin functions,
 * methods, attributes, and language statements, which operate on record
 * buffers.
 * <p>
 * Converted business logic invokes these methods on DMO proxies via a
 * DMO-specific interface in the form <code>&lt;DMO interface&gt;.Buf</code>.
 * Each such interface extends both a custom interface of DMO-specific getter
 * and setter methods, as well as the {@link Buffer} interface, which defines
 * the API for these standard services.
 * <p>
 * Dynamic proxy classes for each DMO type extend this class, so that business
 * logic can invoke these standard services directly as method calls on DMO
 * proxy instances.
 * <p>
 * This class is intentionally package private, since it is not meant to be
 * extended outside of this package (and within this package only by dynamic
 * proxy class), nor instantiated directly.
 * <p>
 * This class implements {@link BufferReference} in order to grant its
 * instance methods access to the {@link RecordBuffer} object associated with
 * each instance of the DMO proxy class which extends this class. This access
 * is necessary for these methods to do their work. Even though {@link
 * #buffer()} is a public method, external packages will not be able to invoke
 * this method (except through reflection). This is by design, as business
 * logic is not meant to have access to <code>RecordBuffer</code> instances.
 * The default implementation of this method here throws an exception, as it
 * is never intended to be called, but it needs to be present in order for
 * this class to compile, since it is referenced by other methods. Each DMO
 * proxy class overrides this method to provide a working implementation.
 */
public abstract class BufferImpl
extends HandleChain
implements Buffer
{
   /** Corresponds to ADM-DATA attribute. */
   private final character admData = new character();

   /** Corresponds to UNIQUE-ID attribute. */
   private final integer uniqueID = new integer(UniqueIdGenerator.getUniqueId(IdKind.BUFFER));

   /** The ERROR property accessible through isError() and setError() methods. */
   private boolean hasError = false;

   /**
    * The default NAMESPACE-URI attribute. Because it is null, it will assigned to BDT as
    * <code>unknown</code> value.
    *
    * TODO:
    * This attribute (NAMESPACE-URI) from CREATE TEMP-TABLE is not handled at all in the P2J
    * conversion so it completely removed in the generated code.
    * It should be stored in the buffer/table.
    * For the moment it is stored here so the getter/setter to partially work.
    */
   private String namespaceURI = null;

   /** The mode for finding (first, last, unique) records in a dynamic call to a buffer handle. */
   private enum FindMode
   {
      FIRST,   /* apply first() to a query */
      LAST,    /* apply first() to a query */
      UNIQUE   /* apply first() to a query */
   }

   /**
    * Worker static method for wrapping ErrorConditionException and reusing
    * in all P2JQuery.findFirst() implementations
    *
    * @param   findQuery
    *          instance of FindQuery
    *
    * @return  <code>true</code> if FIND-FIRST() succeeds,
    *          else <code>false</code>.
    */
   private static logical wrapFirst(FindQuery findQuery, LockType lockType)
   {
      boolean success = true;
      try
      {
         findQuery.first(lockType);
      }
      catch (ErrorConditionException exc)
      {
         success = false;
      }
      return new logical(success);
   }

   /**
    * Worker static method for wrapping ErrorConditionException and reusing
    * in all P2JQuery.findLast() implementations
    *
    * @param   findQuery
    *          instance of FindQuery
    *
    * @return  <code>true</code> if FIND-LAST() succeeds,
    *          else <code>false</code>.
    */
   private static logical wrapLast(FindQuery findQuery, LockType lockType)
   {
      boolean success = true;
      try
      {
         findQuery.last(lockType);
      }
      catch (ErrorConditionException exc)
      {
         success = false;
      }
      return new logical(success);
   }

   /**
    * Worker static method for wrapping ErrorConditionException and reusing
    * in all P2JQuery.findUnique() implementations
    *
    * @param   findQuery
    *          instance of FindQuery
    *
    * @return  <code>true</code> if FIND-UNIQUE() succeeds,
    *          else <code>false</code>.
    */
   private static logical wrapUnique(FindQuery findQuery, LockType lockType)
   {
      boolean success = true;
      try
      {
         findQuery.unique(lockType);
      }
      catch (ErrorConditionException exc)
      {
         success = false;
      }
      return new logical(success);
   }

   /**
    * Checks if transaction is active otherwise records or shows error message.
    *
    * @return  <code>true</code> if a transaction is active else <code>false</code>.
    */
   private static boolean verifyTransaction()
   {
      if (TransactionManager.isTransaction())
      {
         return true;
      }
      else
      {
         String msg = "Unable to CREATE/DELETE unless a TRANSACTION is running";
         ErrorManager.recordOrShowError(7345, msg, false);
         return false;
      }
   }

   /**
   * Get the {@link RecordBuffer} instance associated with this object.
   * <p>
   * This implementation exists only to satisfy the Java compiler. This
   * method is only meaningful when a proxy DMO class is created by {@link
   * RecordBuffer}, which will implement this method to return an instance
   * of <code>RecordBuffer</code> associated with the proxy DMO instance.
   *
   * @return  Record buffer associated with this object.
   *
   * @throws  UnsupportedOperationException
   *          always.
   */
   protected abstract RecordBuffer buffer();
   
   /**
    * Reports if this object is valid for use.  
    *
    * @return  <code>true</code> if we are valid (can be used).
    */
   public boolean valid()
   {
      return buffer().getOpenScopeCount() > 0;
   }
   
   /**
    * Obtain the handle to the buffer that was created by default for the temporary table backing
    * this buffer. Note that this implementation presumes that this is a temp-table buffer. It
    * should not be invoked on a permanent table buffer.
    * 
    * @return  Default buffer for the temporary table backing this buffer, or an unknown handle if
    *          that resource is not valid.
    */
   public handle defaultBufferHandle()
   {
      handle tempTableHandle = ((TemporaryBuffer) buffer()).tableHandle();
      if (!tempTableHandle._isValid())
      {
         return new handle();
      }
      
      TempTable tempTable = (TempTable) tempTableHandle.getResource();
      
      return tempTable.defaultBufferHandle();
   }
   
  /**
   * Get the record ID of the current buffer's current record, if any.  P2J
   * implementation of the RECID function.
   * <p>
   * This method will return a 32-bit recid which represents the internal
   * row ID of the given record.  If the actual row ID (which is implemented
   * as a wider integer) cannot fit due to overflow, an error condition will
   * be raised.
   *
   * @return  Record ID of the underlying data record. If no record currently
   *          backs the buffer associated with <code>dmo</code>, a recid
   *          initialized as unknown value is returned.
   * 
   * @throws  ErrorConditionException
   *          if the actual record ID is too wide to fit within a 32-bit
   *          integer.
   */
  public recid recordID()
  {
     return buffer().recordID();
  }
  
  /**
   * Get the row ID of the current buffer's current record, if any.  P2J
   * implementation of the ROWID function.
   *
   * @return  Row ID of the underlying data record.  If no record currently
   *          backs the buffer associated with <code>dmo</code>, a
   *          <code>rowid</code> initialized as unknown value is returned.
   */
  public rowid rowID()
  {
     return buffer().rowID();
  }
  
   /**
    * Indicate whether a backing data record currently is available.
    *
    * @return  <code>true</code> if data is available, <code>false</code> if
    *          no record backs this buffer.
    */
   public logical available()
   {
      return (new logical(_available()));
   }
   
   /**
    * Indicate whether a backing data record currently is available.
    *
    * @return  <code>true</code> if data is available, <code>false</code> if
    *          no record backs this buffer.
    */
   public boolean _available()
   {
      return buffer().isAvailable();
   }
   
   /**
    * Checks if the record in the buffer has changed in database since the 
    * last <code>FIND CURRENT</code> or <code>GET CURRENT</code>.
    * P2J implementation of 4GL function <code>CURRENT-CHANGED</code>.
    *
    * @return  <code>true</code> if the database record differs from the
    *          value in memory buffer that was queried with the last
    *          <code>FIND</code> and <code>false</code> otherwise
    */
   public logical currentChanged()
   {
      return new logical(_currentChanged());
   }
   
   /**
    * Checks if the record in the buffer has changed in database since the 
    * last <code>FIND CURRENT</code> or <code>GET CURRENT</code>.
    * P2J implementation of 4GL function <code>CURRENT-CHANGED</code>.
    *
    * @return  <code>true</code> if the database record differs from the
    *          value in memory buffer that was queried with the last
    *          <code>FIND</code> and <code>false</code> otherwise
    */
   public boolean _currentChanged()
   {
      return buffer().isRecordChanged();
   }
   
   /**
    * Indicate whether the buffer associated currently contains a record, and
    * whether that record is transient.  A record is transient if it has been
    * newly created and has not yet been saved to the database.
    *
    * @return  <code>true</code> if a record exists and is new, else
    *          <code>false</code>.
    */
   public logical newlyCreated()
   {
      return (new logical(_newlyCreated()));
   }
   
   /**
    * Indicate whether the buffer associated currently contains a record, and
    * whether that record is transient.  A record is transient if it has been
    * newly created and has not yet been saved to the database.
    *
    * @return  <code>true</code> if a record exists and is new, else
    *          <code>false</code>.
    */
   public boolean _newlyCreated()
   {
      return buffer().isNewlyCreated();
   }
   
   /**
    * Indicate whether the last record retrieval attempt was denied because
    * the target record was locked.
    *
    * @return  <code>true</code> if the record was locked;  <code>false</code>
    *          if the attempt succeeded, or if this query has not yet
    *          attempted to retrieve a record.
    */
   public logical locked()
   {
      return (new logical(_locked()));
   }
   
   /**
    * Indicate whether the last record retrieval attempt was denied because
    * the target record was locked.
    *
    * @return  <code>true</code> if the record was locked;  <code>false</code>
    *          if the attempt succeeded, or if this query has not yet
    *          attempted to retrieve a record.
    */
   public boolean _locked()
   {
      return buffer().wasLocked();
   }
   
   /**
    * Indicate whether the last record retrieval attempt failed because the
    * query criteria was ambiguous.
    *
    * @return  <code>true</code> if the query was ambiguous;
    *          <code>false</code> if the attempt succeeded, or if no attempt
    *          has yet been made.
    */
   public logical ambiguous()
   {
      return (new logical(_ambiguous()));
   }
   
   /**
    * Indicate whether the last record retrieval attempt failed because the
    * query criteria was ambiguous.
    *
    * @return  <code>true</code> if the query was ambiguous;
    *          <code>false</code> if the attempt succeeded, or if no attempt
    *          has yet been made.
    */
   public boolean _ambiguous()
   {
      return buffer().wasAmbiguous();
   }
   
   /**
    * Create a new record in the buffer.
    *
    * @throws  ErrorConditionException
    *          if there is an error retrieving the next available primary key
    *          from the database or locking the record.
    */
   public void create()
   {
      buffer().create();
   }
   
   /**
    * Delete the record currently in the buffer, if any.
    *
    * @return  <code>true</code> if the delete was possible.
    *
    * @throws  ErrorConditionException
    *          if no record currently is loaded in the buffer, or if there is
    *          an error deleting the current record from the database.
    */
   public logical deleteRecord()
   {
      RecordBuffer buffer = buffer();

      if (!buffer.isTemporary() && !verifyTransaction())
      {
         return new logical(false);
      }
      try
      {
         buffer.delete();
         return new logical(true);
      }
      catch (PersistenceException exc)
      {
         ErrorManager.recordOrThrowError(exc);

         return new logical(false);
      }
   }
   
   /**
    * Delete the record currently in the buffer, IF and ONLY IF the given
    * validator expression evaluates <code>true</code>.
    *
    * @param   validator
    *          The user-defined callback that is to be tested to determine if the
    *          delete should or should not occur.
    *
    * @throws  ErrorConditionException
    *          if no record currently is loaded in the buffer, or if there is
    *          an error deleting the current record from the database or if the
    *          validation expression fails.
    */
   public logical deleteRecord(BufferValidator validator)
   {
      // TODO: is this the correct behavior?
      if (validator.validate())
      {
         return deleteRecord();
      }
      else
      {
         return new logical(false);
      }
   }
   
   /**
    * Delete all records associated with this buffer, which match the criteria
    * specified by the given where clause.  Note that this where clause must
    * not contain DMO alias qualifiers.
    * <p>
    * <b>NOTE:</b>  the current implementation must not be invoked for a DMO
    * which is backed by a permanent table;  only temp-table bulk delete is
    * supported at this time.
    * 
    * @param   where
    *          An HQL where clause snippet which defines the restriction
    *          criteria to apply to the delete.  All references to properties
    *          in a DMO must be unqualified.
    * @param   args
    *          Query substitution parameters required by the where clause.
    * 
    * @throws  ErrorConditionException
    *          if there is an error deleting the records.
    * @throws  UnsupportedOperationException
    *          if invoked on a buffer associated with a permanent table.
    */
   public void delete(String where, Object ...args)
   {
      try
      {
         buffer().delete(where, args);
      }
      catch (PersistenceException exc)
      {
         ErrorManager.recordOrThrowError(exc);
      }
   }
   
   /**
    * Bulk delete all records associated with the buffer.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the bulk delete operation.
    */
   public logical deleteAll()
   {
      buffer().deleteAll();
      
      // TODO: when does this return false?
      return new logical(true);
   }
   
   /**
    * Aggregate method which averages the values of a single column across rows in the table
    * backing this buffer.
    * 
    * @param   column
    *          An expression indicating the name of the column to aggregate.
    * @param   num
    *          Variable into which to store the aggregate result.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the aggregate operation.
    */
   public void average(String column, NumberType num)
   {
      average(column, num, null);
   }
   
   /**
    * Aggregate method which averages the values of a single column across rows in the table
    * backing this buffer.
    * 
    * @param   column
    *          An expression indicating the name of the column to aggregate.
    * @param   num
    *          Variable into which to store the aggregate result.
    * @param   where
    *          HQL where clause used to filter the records.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the aggregate operation.
    */
   public void average(String column, NumberType num, String where)
   {
      average(column, num, where, (Object[]) null);
   }
   
   /**
    * Aggregate method which averages the values of a single column across rows in the table
    * backing this buffer.
    * 
    * @param   column
    *          An expression indicating the name of the column to aggregate.
    * @param   num
    *          Variable into which to store the aggregate result.
    * @param   where
    *          HQL where clause used to filter the records.
    * @param   args
    *          Query substitution arguments.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the aggregate operation.
    */
   public void average(String column, NumberType num, String where, Object...args)
   {
      buffer().average(column, num, where, args);
   }
   
   /**
    * Aggregate method which counts the rows in the table backing this buffer.
    * 
    * @param   column
    *          An expression indicating the name of the column to aggregate. May be the wildcard
    *          character ("*").
    * @param   num
    *          Variable into which to store the aggregate result.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the aggregate operation.
    */
   public void count(String column, NumberType num)
   {
      count(column, num, null);
   }
   
   /**
    * Aggregate method which counts the rows in the table backing this buffer.
    * 
    * @param   column
    *          An expression indicating the name of the column to aggregate. May be the wildcard
    *          character ("*").
    * @param   num
    *          Variable into which to store the aggregate result.
    * @param   where
    *          HQL where clause used to filter the records.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the aggregate operation.
    */
   public void count(String column, NumberType num, String where)
   {
      count(column, num, where, (Object[]) null);
   }
   
   /**
    * Aggregate method which counts the rows in the table backing this buffer.
    * 
    * @param   column
    *          An expression indicating the name of the column to aggregate. May be the wildcard
    *          character ("*").
    * @param   num
    *          Variable into which to store the aggregate result.
    * @param   where
    *          HQL where clause used to filter the records.
    * @param   args
    *          Query substitution arguments.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the aggregate operation.
    */
   public void count(String column, NumberType num, String where, Object...args)
   {
      buffer().count(column, num, where, args);
   }
   
   /**
    * Aggregate method which finds the minimum value of a single column across rows in the table
    * backing this buffer.
    * 
    * @param   column
    *          An expression indicating the name of the column to aggregate.
    * @param   num
    *          Variable into which to store the aggregate result.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the aggregate operation.
    */
   public void minimum(String column, NumberType num)
   {
      minimum(column, num, null);
   }
   
   /**
    * Aggregate method which finds the minimum value of a single column across rows in the table
    * backing this buffer.
    * 
    * @param   column
    *          An expression indicating the name of the column to aggregate.
    * @param   num
    *          Variable into which to store the aggregate result.
    * @param   where
    *          HQL where clause used to filter the records.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the aggregate operation.
    */
   public void minimum(String column, NumberType num, String where)
   {
      minimum(column, num, where, (Object[]) null);
   }
   
   /**
    * Aggregate method which finds the minimum value of a single column across rows in the table
    * backing this buffer.
    * 
    * @param   column
    *          An expression indicating the name of the column to aggregate.
    * @param   num
    *          Variable into which to store the aggregate result.
    * @param   where
    *          HQL where clause used to filter the records.
    * @param   args
    *          Query substitution arguments.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the aggregate operation.
    */
   public void minimum(String column, NumberType num, String where, Object...args)
   {
      buffer().minimum(column, num, where, args);
   }
   
   /**
    * Aggregate method which finds the maximum value of a single column across rows in the table
    * backing this buffer.
    * 
    * @param   column
    *          An expression indicating the name of the column to aggregate.
    * @param   num
    *          Variable into which to store the aggregate result.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the aggregate operation.
    */
   public void maximum(String column, NumberType num)
   {
      maximum(column, num, null);
   }
   
   /**
    * Aggregate method which finds the maximum value of a single column across rows in the table
    * backing this buffer.
    * 
    * @param   column
    *          An expression indicating the name of the column to aggregate.
    * @param   num
    *          Variable into which to store the aggregate result.
    * @param   where
    *          HQL where clause used to filter the records.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the aggregate operation.
    */
   public void maximum(String column, NumberType num, String where)
   {
      maximum(column, num, where, (Object[]) null);
   }
   
   /**
    * Aggregate method which finds the maximum value of a single column across rows in the table
    * backing this buffer.
    * 
    * @param   column
    *          An expression indicating the name of the column to aggregate.
    * @param   num
    *          Variable into which to store the aggregate result.
    * @param   where
    *          HQL where clause used to filter the records.
    * @param   args
    *          Query substitution arguments.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the aggregate operation.
    */
   public void maximum(String column, NumberType num, String where, Object...args)
   {
      buffer().maximum(column, num, where, args);
   }
   
   /**
    * Aggregate method which sums the values of a single column across rows in the table
    * backing this buffer.
    * 
    * @param   column
    *          An expression indicating the name of the column to aggregate.
    * @param   num
    *          Variable into which to store the aggregate result.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the aggregate operation.
    */
   public void sum(String column, NumberType num)
   {
      sum(column, num, null);
   }
   
   /**
    * Aggregate method which sums the values of a single column across rows in the table
    * backing this buffer.
    * 
    * @param   column
    *          An expression indicating the name of the column to aggregate.
    * @param   num
    *          Variable into which to store the aggregate result.
    * @param   where
    *          HQL where clause used to filter the records.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the aggregate operation.
    */
   public void sum(String column, NumberType num, String where)
   {
      sum(column, num, where, (Object[]) null);
   }
   
   /**
    * Aggregate method which sums the values of a single column across rows in the table
    * backing this buffer.
    * 
    * @param   column
    *          An expression indicating the name of the column to aggregate.
    * @param   num
    *          Variable into which to store the aggregate result.
    * @param   where
    *          HQL where clause used to filter the records.
    * @param   args
    *          Query substitution arguments.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the aggregate operation.
    */
   public void sum(String column, NumberType num, String where, Object...args)
   {
      buffer().sum(column, num, where, args);
   }
   
   /**
    * Validate and release the record currently in the buffer, if any.
    *
    * @throws  ErrorConditionException
    *          if the buffer fails validation.
    */
   public void release()
   {
      buffer().release();
   }
   
   /**
    * Validate the record currently in the buffer, if any.
    *
    * @throws  ErrorConditionException
    *          if the buffer fails validation.
    */
   public void validate()
   {
      buffer().validate();
   }

   /**
    * Conversion of BUFFER-FIELD() method (KW_BUF_FLD).
    *
    * Gets the handle to a particular buffer of a query or ProDataSet object.
    *
    * @param   fieldNumber
    *          field number. An INTEGER expression representing the sequence
    *          number of the field in the buffer.
    *
    * @return  handle to a particular field in the buffer.
    */
   @Override
   public handle bufferField(int fieldNumber)
   {
      final int fieldCount = _numFields();
      if (fieldNumber <= 0 || fieldNumber > fieldCount)
      {
         ErrorManager.recordOrShowError(7350,
                "BUFFER-FIELD sequence number must be from 1 to " +
                      fieldCount + " for buffer " + this.name.getValue(),
                false, false, false);
         return (handle) BaseDataType.generateUnknown(handle.class);
      }

      String hibernateName;

      TempTable parentTempTable = buffer().getParentTable();
      if (parentTempTable != null && parentTempTable._dynamic())
      {
         // dynamic table
         hibernateName = DynamicTablesHelper.getHibernateFieldName(buffer().getDMOInterface(),
                                                                   fieldNumber);
      }
      else
      {
         // static temporary or permanent table
         Database database = buffer().getDatabase();
         if (database.equals(DatabaseManager.TEMP_TABLE_DB))
         {
            hibernateName = TableMapper.getPropertyName(parentTempTable, fieldNumber);
         }
         else
         {
            String schema = DatabaseManager.getSchema(database);
            String legacyTableName = TableMapper.getLegacyName(this);
            hibernateName = TableMapper.getPropertyName(schema, legacyTableName, fieldNumber);
         }
      }

      return new handle(new FieldReference((DataModelObject) this, hibernateName));
   }

   /**
    * Implementation of BUFFER-FIELD() method (KW_BUF_FLD).
    *
    * Gets the handle to a particular field within a buffer.
    *
    * @param   fieldName
    *          field name. A CHARACTER string expression representing the name
    *          of the field in the buffer.
    *
    * @return  handle to a particular field in the buffer.
    */
   @Override
   public handle bufferField(String fieldName)
   {
      String hibernateName;
      RecordBuffer buffer = buffer();
      String legacyTableName = TableMapper.getLegacyName(this);
      
      if (buffer.isTemporary())
      {
         TempTable parentTempTable = buffer.getParentTable();
         if (parentTempTable._dynamic())
         {
            hibernateName = DynamicTablesHelper.getHibernateFieldName(buffer.getDMOInterface(),
                                                                      fieldName);
         }
         else
         {
            hibernateName = TableMapper.getPropertyName(parentTempTable, fieldName);
         }
      }
      else
      {
         Database database = buffer.getDatabase();
         String schema = DatabaseManager.getSchema(database);
         hibernateName = TableMapper.getPropertyName(schema, legacyTableName, fieldName);
      }
      
      if (hibernateName == null)
      {
         String msg = String.format("BUFFER-FIELD %s was not found in buffer %s",
                                    fieldName,
                                    legacyTableName);
         ErrorManager.recordOrThrowError(7351, msg);
         
         // return unknown handle if in silent error mode
         return (new handle());
      }
      
      return (new handle(new FieldReference((DataModelObject) this, hibernateName)));
   }

   /**
    * Conversion of BUFFER-COPY() method (KW_BUF_COPY).
    *
    * This method copies any common fields, determined by name, data type,
    * and extent-matching, from the source buffer to the receiving buffer.
    * If there are fields in one buffer that do not exist in the other, they
    * are ignored. This method is used to accommodate temp-tables of joins.
    * 
    * @param   buffer
    *          The source buffer to copy from.
    *
    * @return  <code>true</code> if BUFFER-COPY() succeeds, else <code>false</code>.
    */
   public logical bufferCopy(Buffer buffer)
   {
      return bufferCopy(buffer, null, null);
   }

   /**
    * Conversion of BUFFER-COPY() method (KW_BUF_COPY).
    *
    * This method copies any common fields, determined by name, data type,
    * and extent-matching, from the source buffer to the receiving buffer.
    * If there are fields in one buffer that do not exist in the other, they
    * are ignored. This method is used to accommodate temp-tables of joins.
    *
    * @param   buffer
    *          The source buffer to copy from.
    * @param   except
    *          A comma-separated list of fields that will be ignored in the copy process.
    *
    * @return  <code>true</code> if BUFFER-COPY() succeeds, else <code>false</code>.
    */
   public logical bufferCopy(Buffer buffer, character except)
   {
      return bufferCopy(buffer, except, null);
   }

   /**
    * Conversion of BUFFER-COPY() method (KW_BUF_COPY).
    *
    * This method copies any common fields, determined by name, data type,
    * and extent-matching, from the source buffer to the receiving buffer.
    * If there are fields in one buffer that do not exist in the other, they
    * are ignored. This method is used to accommodate temp-tables of joins.
    * 
    * @param   buffer
    *          The source buffer to copy from.
    * @param   except
    *          A comma-separated list of fields that will be ignored in the copy process.
    * @param   pairs
    *          A comma-separated list of field-name pairs to be copied.  Each pair must contain 
    *          one field name from the source and one field name from the target.
    * 
    * @return  <code>true</code> if BUFFER-COPY() succeeds, else <code>false</code>.
    */
   public logical bufferCopy(Buffer buffer, character except, character pairs)
   {
      return RecordBuffer.copy(buffer(), ((BufferImpl) buffer).buffer(), except, pairs);
   }

   /**
    * Conversion of BUFFER-COPY() method (KW_BUF_COPY).
    *
    * This method copies any common fields, determined by name, data type,
    * and extent-matching, from the source buffer to the receiving buffer.
    * If there are fields in one buffer that do not exist in the other, they
    * are ignored. This method is used to accommodate temp-tables of joins.
    * 
    * @param   buffer
    *          The source buffer to copy from.
    * @param   except
    *          A comma-separated list of fields that will be ignored in the copy process.
    * @param   pairs
    *          A comma-separated list of field-name pairs to be copied.  Each pair must contain 
    *          one field name from the source and one field name from the target.
    * @param   noLobs
    *          If <code>TRUE</code>, BLOB and CLOB fields are ignored during the copy.
    * 
    * @return  <code>true</code> if BUFFER-COPY() succeeds, else <code>false</code>.
    */
   public logical bufferCopy(Buffer buffer, character except, character pairs, logical noLobs)
   {
      throw new UnsupportedOperationException();
   }

   /**
    * Conversion of BUFFER-COPY() method (KW_BUF_COPY).
    *
    * This method copies any common fields, determined by name, data type,
    * and extent-matching, from the source buffer to the receiving buffer.
    * If there are fields in one buffer that do not exist in the other, they
    * are ignored. This method is used to accommodate temp-tables of joins.
    * 
    * @param   bufHandle
    *          The handle of the source buffer to copy from.
    * 
    * @return  <code>true</code> if BUFFER-COPY() succeeds, else <code>false</code>.
    */
   public logical bufferCopy(handle bufHandle)
   {
      return bufferCopy((Buffer) bufHandle.get());
   }

   /**
    * Conversion of BUFFER-COPY() method (KW_BUF_COPY).
    *
    * This method copies any common fields, determined by name, data type,
    * and extent-matching, from the source buffer to the receiving buffer.
    * If there are fields in one buffer that do not exist in the other, they
    * are ignored. This method is used to accommodate temp-tables of joins.
    * 
    * @param   bufHandle
    *          The handle of the source buffer to copy from.
    * @param   except
    *          A comma-separated list of fields that will be ignored in the copy process.
    * 
    * @return  <code>true</code> if BUFFER-COPY() succeeds, else <code>false</code>.
    */
   public logical bufferCopy(handle bufHandle, character except)
   {
      return bufferCopy((Buffer) bufHandle.get(), except);
   }

   /**
    * Conversion of BUFFER-COPY() method (KW_BUF_COPY).
    *
    * This method copies any common fields, determined by name, data type,
    * and extent-matching, from the source buffer to the receiving buffer.
    * If there are fields in one buffer that do not exist in the other, they
    * are ignored. This method is used to accommodate temp-tables of joins.
    * 
    * @param   bufHandle
    *          The handle of the source buffer to copy from.
    * @param   except
    *          A comma-separated list of fields that will be ignored in the copy process.
    * @param   pairs
    *          A comma-separated list of field-name pairs to be copied.  Each pair must contain 
    *          one field name from the source and one field name from the target.
    * 
    * @return  <code>true</code> if BUFFER-COPY() succeeds, else <code>false</code>.
    */
   public logical bufferCopy(handle bufHandle, character except, character pairs)   {
      return bufferCopy((Buffer) bufHandle.get(), except, pairs);
   }

   /**
    * Conversion of BUFFER-COPY() method (KW_BUF_COPY).
    *
    * This method copies any common fields, determined by name, data type,
    * and extent-matching, from the source buffer to the receiving buffer.
    * If there are fields in one buffer that do not exist in the other, they
    * are ignored. This method is used to accommodate temp-tables of joins.
    * 
    * @param   bufHandle
    *          The handle of the source buffer to copy from.
    * @param   except
    *          A comma-separated list of fields that will be ignored in the copy process.
    * @param   pairs
    *          A comma-separated list of field-name pairs to be copied.  Each pair must contain 
    *          one field name from the source and one field name from the target.
    * @param   noLobs
    *          If <code>TRUE</code>, BLOB and CLOB fields are ignored during the copy.
    * 
    * @return  <code>true</code> if BUFFER-COPY() succeeds, else <code>false</code>.
    */
   public logical bufferCopy(handle bufHandle,
                             character except,
                             character pairs,
                             logical noLobs)
   {
      throw new UnsupportedOperationException();
   }

   /**
    * Conversion of BUFFER-COMPARE() method (KW_BUF_COMP).
    *
    * This method does a rough compare of any common fields, determined by
    * name, data type, and extent-matching, between the source buffer and the
    * target buffer. The resulting logical value is either TRUE or FALSE as
    * a whole. A single field that does not compare causes the entire buffer
    * to return FALSE. If there are fields in one buffer that do not exist in
    * the other, they are ignored.
    *
    * @param   buffer
    *          Buffer with which to compare.
    *
    * @return  <code>true</code> if BUFFER-COMPARE() succeeds,
    *          else <code>false</code>.
    */
   public logical bufferCompare(Buffer buffer)
   {
      return bufferCompare(buffer, null, null, null);
   }

   /**
    * Conversion of BUFFER-COMPARE() method (KW_BUF_COMP).
    *
    * This method does a rough compare of any common fields, determined by
    * name, data type, and extent-matching, between the source buffer and the
    * target buffer. The resulting logical value is either TRUE or FALSE as
    * a whole. A single field that does not compare causes the entire buffer
    * to return FALSE. If there are fields in one buffer that do not exist in
    * the other, they are ignored.
    *
    * @param   buffer
    *          Buffer with which to compare.
    * @param   mode
    *          The text "binary" or "case-sensitive", to provide that type of comparison.
    *
    * @return  <code>true</code> if BUFFER-COMPARE() succeeds,
    *          else <code>false</code>.
    */
   public logical bufferCompare(Buffer buffer, character mode)
   {
      return bufferCompare(buffer, mode, null, null);
   }

   /**
    * Conversion of BUFFER-COMPARE() method (KW_BUF_COMP).
    *
    * This method does a rough compare of any common fields, determined by
    * name, data type, and extent-matching, between the source buffer and the
    * target buffer. The resulting logical value is either TRUE or FALSE as
    * a whole. A single field that does not compare causes the entire buffer
    * to return FALSE. If there are fields in one buffer that do not exist in
    * the other, they are ignored.
    *
    * @param   buffer
    *          Buffer with which to compare.
    * @param   mode
    *          The text "binary" or "case-sensitive", to provide that type of comparison.
    * @param   except
    *          A comma-separated list of fields that will be ignored in the compare process.
    *
    * @return  <code>true</code> if BUFFER-COMPARE() succeeds,
    *          else <code>false</code>.
    */
   public logical bufferCompare(Buffer buffer, character mode, character except)
   {
      return bufferCompare(buffer, mode, except, null);
   }

   /**
    * Conversion of BUFFER-COMPARE() method (KW_BUF_COMP).
    *
    * This method does a rough compare of any common fields, determined by
    * name, data type, and extent-matching, between the source buffer and the
    * target buffer. The resulting logical value is either TRUE or FALSE as
    * a whole. A single field that does not compare causes the entire buffer
    * to return FALSE. If there are fields in one buffer that do not exist in
    * the other, they are ignored.
    *
    * @param   srcBuffer
    *          Buffer with which to compare.
    * @param   mode
    *          The text "binary" or "case-sensitive", to provide that type of comparison.
    * @param   except
    *          A comma-separated list of fields that will be ignored in the compare process.
    * @param   pairs
    *          A comma-separated list of field-name pairs to be compared.  Each pair must
    *          contain one field name from the source and one field name from the target.
    *
    * @return  <code>true</code> if BUFFER-COMPARE() succeeds,
    *          else <code>false</code>.
    */
   public logical bufferCompare(Buffer srcBuffer,
                                character mode,
                                character except,
                                character pairs)
   {
      return RecordBuffer.compare(((BufferImpl) srcBuffer).buffer(),
                                  buffer(),
                                  mode,
                                  except,
                                  pairs);
   }

   /**
    * Conversion of BUFFER-COMPARE() method (KW_BUF_COMP).
    *
    * This method does a rough compare of any common fields, determined by
    * name, data type, and extent-matching, between the source buffer and the
    * target buffer. The resulting logical value is either TRUE or FALSE as
    * a whole. A single field that does not compare causes the entire buffer
    * to return FALSE. If there are fields in one buffer that do not exist in
    * the other, they are ignored.
    *
    * @param   buffer
    *          Buffer with which to compare.
    * @param   mode
    *          The text "binary" or "case-sensitive", to provide that type of comparison.
    * @param   except
    *          A comma-separated list of fields that will be ignored in the compare process.
    * @param   pairs
    *          A comma-separated list of field-name pairs to be compared.  Each pair must
    *          contain one field name from the source and one field name from the target.
    * @param   noLobs
    *          If <code>TRUE</code>, BLOB and CLOB fields are ignored during the copy.
    *
    * @return  <code>true</code> if BUFFER-COMPARE() succeeds,
    *          else <code>false</code>.
    */
   public logical bufferCompare(Buffer    buffer,
                                character mode,
                                character except,
                                character pairs,
                                logical   noLobs)
   {
      throw new UnsupportedOperationException();
   }

   /**
    * Conversion of BUFFER-COMPARE() method (KW_BUF_COMP).
    *
    * This method does a rough compare of any common fields, determined by
    * name, data type, and extent-matching, between the source buffer and the
    * target buffer. The resulting logical value is either TRUE or FALSE as
    * a whole. A single field that does not compare causes the entire buffer
    * to return FALSE. If there are fields in one buffer that do not exist in
    * the other, they are ignored.
    *
    * @param   bufHandle
    *          Handle of buffer with which to compare.
    *
    * @return  <code>true</code> if BUFFER-COMPARE() succeeds,
    *          else <code>false</code>.
    */
   public logical bufferCompare(handle bufHandle)
   {
      return bufferCompare((Buffer) bufHandle.get());
   }

   /**
    * Conversion of BUFFER-COMPARE() method (KW_BUF_COMP).
    *
    * This method does a rough compare of any common fields, determined by
    * name, data type, and extent-matching, between the source buffer and the
    * target buffer. The resulting logical value is either TRUE or FALSE as
    * a whole. A single field that does not compare causes the entire buffer
    * to return FALSE. If there are fields in one buffer that do not exist in
    * the other, they are ignored.
    *
    * @param   bufHandle
    *          Handle of buffer with which to compare.
    * @param   mode
    *          The text "binary" or "case-sensitive", to provide that type of comparison.
    *
    * @return  <code>true</code> if BUFFER-COMPARE() succeeds,
    *          else <code>false</code>.
    */
   public logical bufferCompare(handle bufHandle, character mode)
   {
      return bufferCompare((Buffer) bufHandle.get(), mode);
   }

   /**
    * Conversion of BUFFER-COMPARE() method (KW_BUF_COMP).
    *
    * This method does a rough compare of any common fields, determined by
    * name, data type, and extent-matching, between the source buffer and the
    * target buffer. The resulting logical value is either TRUE or FALSE as
    * a whole. A single field that does not compare causes the entire buffer
    * to return FALSE. If there are fields in one buffer that do not exist in
    * the other, they are ignored.
    *
    * @param   bufHandle
    *          Handle of buffer with which to compare.
    * @param   mode
    *          The text "binary" or "case-sensitive", to provide that type of comparison.
    * @param   except
    *          A comma-separated list of fields that will be ignored in the compare process.
    *
    * @return  <code>true</code> if BUFFER-COMPARE() succeeds,
    *          else <code>false</code>.
    */
   public logical bufferCompare(handle bufHandle, character mode, character except)
   {
      return bufferCompare((Buffer) bufHandle.get(), mode, except);
   }

   /**
    * Conversion of BUFFER-COMPARE() method (KW_BUF_COMP).
    *
    * This method does a rough compare of any common fields, determined by
    * name, data type, and extent-matching, between the source buffer and the
    * target buffer. The resulting logical value is either TRUE or FALSE as
    * a whole. A single field that does not compare causes the entire buffer
    * to return FALSE. If there are fields in one buffer that do not exist in
    * the other, they are ignored.
    *
    * @param   bufHandle
    *          Handle of buffer with which to compare.
    * @param   mode
    *          The text "binary" or "case-sensitive", to provide that type of comparison.
    * @param   except
    *          A comma-separated list of fields that will be ignored in the compare process.
    * @param   pairs
    *          A comma-separated list of field-name pairs to be compared.  Each pair must
    *          contain one field name from the source and one field name from the target.
    *
    * @return  <code>true</code> if BUFFER-COMPARE() succeeds,
    *          else <code>false</code>.
    */
   public logical bufferCompare(handle bufHandle,
                                character mode,
                                character except,
                                character pairs)
   {
      return bufferCompare((Buffer) bufHandle.get(), mode, except, pairs);
   }

   /**
    * Conversion of BUFFER-COMPARE() method (KW_BUF_COMP).
    *
    * This method does a rough compare of any common fields, determined by
    * name, data type, and extent-matching, between the source buffer and the
    * target buffer. The resulting logical value is either TRUE or FALSE as
    * a whole. A single field that does not compare causes the entire buffer
    * to return FALSE. If there are fields in one buffer that do not exist in
    * the other, they are ignored.
    *
    * @param   bufHandle
    *          Handle of buffer with which to compare.
    * @param   mode
    *          The text "binary" or "case-sensitive", to provide that type of comparison.
    * @param   except
    *          A comma-separated list of fields that will be ignored in the compare process.
    * @param   pairs
    *          A comma-separated list of field-name pairs to be compared.  Each pair must
    *          contain one field name from the source and one field name from the target.
    * @param   noLobs
    *          If <code>TRUE</code>, BLOB and CLOB fields are ignored during the compare.
    *
    * @return  <code>true</code> if BUFFER-COMPARE() succeeds,
    *          else <code>false</code>.
    */
   public logical bufferCompare(handle    bufHandle,
                                character mode,
                                character except,
                                character pairs,
                                logical   noLobs)
   {
      throw new UnsupportedOperationException();
   }

   /**
    * Conversion of BUFFER-CREATE() method (KW_BUF_CREA).
    *
    * Creates a record, sets fields to their default values, and moves a copy of the record into 
    * the buffer.
    *
    * @return  <code>true</code> if successful
    */
   public logical bufferCreate()
   {
      RecordBuffer buffer = buffer();
      if (!buffer.isTemporary() && !verifyTransaction())
      {
         return new logical(false);
      }

      boolean success = true;
      try
      {
         buffer.create();
      }
      catch (ErrorConditionException exc)
      {
         success = false;
      }
      return new logical(success);
   }

   /**
    * Conversion of BUFFER-RELEASE() method (KW_BUF_REL).
    *
    * Releases a record from a buffer object. The BUFFER-RELEASE method
    * corresponds to the RELEASE statement.
    *
    * @return  <code>true</code> if BUFFER-RELEASE() succeeds,
    *          else <code>false</code>.
    */
   public logical bufferRelease()
   {
      boolean success = true;
      try
      {
         buffer().release();
      }
      catch (ErrorConditionException exc)
      {
         success = false;
      }
      return new logical(success);
   }

   /**
    * Dereference a named member from this collection.
    *  
    * @param   memberName
    *          The name of the member to be extracted.
    *
    * @return  The value of the member or unknown (?) if there is no such member.
    */
   @Override
   public BaseDataType dereference(String memberName)
   {
      String hibernateName = _getHibernatePropertyName(memberName);

      if (hibernateName == null)
      {
         String errMsg  = "BUFFER-FIELD " + memberName + " was not found in buffer " +
                          buffer().getTable();
         ErrorManager.recordOrThrowError(7351, errMsg);
         return new unknown();
      }

      return new FieldReference((DataModelObject) this, hibernateName).value();
   }

   /**
    * This overloaded method is used when the dereference is the left-side of an assignment.
    * The implementation will extract the member from the collection and assign it the new value.
    *
    * @param   memberName
    *          The name of the member to be extracted.
    * @param   value
    *          The new value that will be assigned to the extracted field.
    */
   @Override
   public void dereference(String memberName, Object value)
   {
      String hibernateName = _getHibernatePropertyName(memberName);

      if (hibernateName == null)
      {
         String errMsg  = "BUFFER-FIELD " + memberName + " was not found in buffer " +
               buffer().getTable();
         ErrorManager.recordOrThrowError(7351, errMsg);
         return;
      }

      new FieldReference((DataModelObject) this, hibernateName).changeValue(value);
   }

   /**
    * Conversion of FIND-BY-ROWID() method (KW_FIND_BR).
    *
    * Locates the record with the rowid you specify, then moves the record
    * into the buffer.
    *
    * @param   id
    *          An expression of type ROWID that represents the rowid of the
    *          desired record.
    *
    * @return  <code>true</code> if it finds the record, else
    *          <code>false</code>.
    */
   public logical findByRowID(rowid id)
   {
      return findByRowID(id, LockType.LT_SHARE);
   }

   /**
    * Conversion of FIND-BY-ROWID() method (KW_FIND_BR).
    *
    * Locates the record with the rowid you specify, then moves the record into the buffer.
    *
    * @param   id
    *          An expression of type ROWID that represents the rowid of the desired record.
    * @param   lockType
    *          Lock type
    *
    * @return  <code>true</code> if it finds the record, else <code>false</code>.
    */
   public logical findByRowID(rowid id, LockType lockType)
   {
      final RecordBuffer buffer = buffer();
      final String dmoAlias = buffer.getDMOAlias();

      if (id.isUnknown())
      {
         String msg = "Invalid rowid for FIND-BY-ROWID method of BUFFER object " + dmoAlias;
         ErrorManager.recordOrShowError(7341, msg, false);
         return new logical(false);
      }

      try
      {
         ErrorManager.silentErrorEnable();
         buffer.initialize();

         String sort = null;
         TempTable parentTempTable = buffer.getParentTable();
         if (parentTempTable != null)
         {
            // manufacture a rowid sort clause for temp-tables
            sort = dmoAlias + ".id asc";
         }
         else
         {
            IndexHelper idxHelper = IndexHelper.get(buffer.getDatabase());
            try
            {
               P2JIndex idx = idxHelper.getPrimaryIndex(
                     buffer.getDMOImplementationClass().getCanonicalName());
               sort = idx.getOrderByClause(buffer.getDatabase(),
                                           dmoAlias,
                                           buffer.getDMOImplementationClass());
            }
            catch (IllegalStateException | PersistenceException ex)
            {
               // sort stays null
            }
         }

         FindQuery findQuery = new FindQuery(buffer,
                                             dmoAlias + ".id = ?",
                                             null,
                                             sort,
                                             new Object[] { id },
                                             lockType);
         findQuery.unique();

         ErrorManager.silentErrorDisable();

         return new logical(!ErrorManager.isErrorFlag());
      }
      catch (IllegalArgumentException | ErrorConditionException | QueryOffEndException exc)
      {
         // all ignored
      }
      return new logical(false);
   }

   /**
    * Conversion of FIND-BY-ROWID() method (KW_FIND_BR).
    *
    * Locates the record with the rowid you specify, then moves the record
    * into the buffer.
    *
    * @param   id
    *          An expression of type ROWID that represents the rowid of the
    *          desired record.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if it finds the record, else
    *          <code>false</code>.
    *
    * @throws  ErrorConditionException
    */
   public logical findByRowID(rowid id, long lock)
   {
      LockType lockType = fromLockNowait(lock, 0);
      return lockType == null ? new logical(false) : findByRowID(id, lockType);
   }

   /**
    * Conversion of FIND-BY-ROWID() method (KW_FIND_BR).
    *
    * Locates the record with the rowid you specify, then moves the record
    * into the buffer.
    *
    * @param   id
    *          An expression of type ROWID that represents the rowid of the
    *          desired record.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if it finds the record, else
    *          <code>false</code>.
    *
    * @throws  ErrorConditionException
    */
   public logical findByRowID(rowid id, NumberType lock)
   {
      LockType lockType = fromLockNowait(lock.longValue(), 0);
      return lockType == null ? new logical(false) : findByRowID(id, lockType);
   }

   /**
    * Conversion of FIND-BY-ROWID() method (KW_FIND_BR).
    *
    * Locates the record with the rowid you specify, then moves the record
    * into the buffer.
    *
    * @param   id
    *          An expression of type ROWID that represents the rowid of the
    *          desired record.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if it finds the record, else
    *          <code>false</code>.
    *
    * @throws  ErrorConditionException
    */
   public logical findByRowID(rowid id, long lock, long nowait)
   {
      LockType lockType = fromLockNowait(lock, nowait);
      return lockType == null ? new logical(false) : findByRowID(id, lockType);
   }

   /**
    * Conversion of FIND-BY-ROWID() method (KW_FIND_BR).
    *
    * Locates the record with the rowid you specify, then moves the record
    * into the buffer.
    *
    * @param   id
    *          An expression of type ROWID that represents the rowid of the
    *          desired record.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if it finds the record, else
    *          <code>false</code>.
    *
    * @throws  ErrorConditionException
    */
   public logical findByRowID(rowid id, NumberType lock, long nowait)
   {
      LockType lockType = fromLockNowait(lock.longValue(), nowait);
      return lockType == null ? new logical(false) : findByRowID(id, lockType);
   }

   /**
    * Conversion of FIND-BY-ROWID() method (KW_FIND_BR).
    *
    * Locates the record with the rowid you specify, then moves the record
    * into the buffer.
    *
    * @param   id
    *          An expression of type ROWID that represents the rowid of the
    *          desired record.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if it finds the record, else
    *          <code>false</code>.
    *
    * @throws  ErrorConditionException
    */
   public logical findByRowID(rowid id, long lock, NumberType nowait)
   {
      LockType lockType = fromLockNowait(lock, nowait.longValue());
      return lockType == null ? new logical(false) : findByRowID(id, lockType);
   }

   /**
    * Conversion of FIND-BY-ROWID() method (KW_FIND_BR).
    *
    * Locates the record with the rowid you specify, then moves the record
    * into the buffer.
    *
    * @param   id
    *          An expression of type ROWID that represents the rowid of the
    *          desired record.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if it finds the record, else
    *          <code>false</code>.
    *
    * @throws  ErrorConditionException
    */
   public logical findByRowID(rowid id, NumberType lock, NumberType nowait)
   {
      LockType lockType = fromLockNowait(lock.longValue(), nowait.longValue());
      return lockType == null ? new logical(false) : findByRowID(id, lockType);
   }

   /**
    * Conversion of FIND-BY-ROWID() method (KW_FIND_BR).
    *
    * Locates the record with the rowid you specify, then moves the record
    * into the buffer.
    *
    * @param   id
    *          An expression of type ROWID that represents the rowid of the
    *          desired record.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if it finds the record, else
    *          <code>false</code>.
    *
    * @throws  ErrorConditionException
    */
   public logical findByRowID(rowid id, integer lock, int nowait)
   {
      LockType lockType = fromLockNowait(lock.longValue(), nowait);
      return lockType == null ? new logical(false) : findByRowID(id, lockType);
   }

   /**
    * Conversion of FIND-FIRST() method (KW_FIND_1ST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @return  <code>true</code> if FIND-FIRST() succeeds,
    *          else <code>false</code>.
    */
   public logical findFirst()
   {
      return _find(null, FindMode.FIRST, LockType.SHARE);
   }

   /**
    * Conversion of FIND-FIRST() method (KW_FIND_1ST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   predicate
    *          Predicate expression for find.
    *
    * @return  <code>true</code> if FIND-FIRST() succeeds, else <code>false</code>.
    */
   public logical findFirst(character predicate)
   {
      // toJavaType() returns null if predicate is unknown
      return _find(predicate.toJavaType(), FindMode.FIRST, LockType.SHARE);
   }

   /**
    * Conversion of FIND-FIRST() method (KW_FIND_1ST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   predicate
    *          Predicate expression for find.
    * @param   lockType
    *          Lock type
    *
    * @return  <code>true</code> if FIND-FIRST() succeeds,
    *          else <code>false</code>.
    */
   public logical findFirst(character predicate, LockType lockType)
   {
      // toJavaType() returns null if predicate is unknown
      return _find(predicate.toJavaType(), FindMode.FIRST, lockType);
   }

   /**
    * Conversion of FIND-FIRST() method (KW_FIND_1ST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   predicate
    *          Predicate expression for find.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if FIND-FIRST() succeeds, else <code>false</code>.
    */
   public logical findFirst(character predicate, long lock)
   {
      // toJavaType() returns null if predicate is unknown
      return _find(predicate.toJavaType(), FindMode.FIRST, LockType.fromInt((int) lock));
   }

   /**
    * Conversion of FIND-FIRST() method (KW_FIND_1ST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   predicate
    *          Predicate expression for find.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if FIND-FIRST() succeeds, else <code>false</code>.
    */
   public logical findFirst(character predicate, NumberType lock)
   {
      LockType lockType = lock.isUnknown() ? LockType.SHARE : LockType.fromInt(lock.intValue());

      // toJavaType() returns null if predicate is unknown
      return _find(predicate.toJavaType(), FindMode.FIRST, lockType);
   }

   /**
    * Conversion of FIND-FIRST() method (KW_FIND_1ST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   predicate
    *          Predicate expression for find.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          Integer value of one of the following: NO-WAIT, 0, or the Unknown value (?).
    *
    * @return  <code>true</code> if FIND-FIRST() succeeds, else <code>false</code>.
    *
    * @throws  ErrorConditionException
    */
   public logical findFirst(character predicate, long lock, long nowait)
   {
      LockType lockType = fromLockNowait(lock, nowait);
      // toJavaType() returns null if predicate is unknown
      return lockType == null ?
            new logical(false) :
            _find(predicate.toJavaType(), FindMode.FIRST, lockType);
   }

   /**
    * Conversion of FIND-FIRST() method (KW_FIND_1ST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   predicate
    *          Predicate expression for find.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          Integer value of one of the following: NO-WAIT, 0, or the Unknown value (?).
    *
    * @return  <code>true</code> if FIND-FIRST() succeeds, else <code>false</code>.
    *
    * @throws  ErrorConditionException
    */
   public logical findFirst(character predicate, NumberType lock, long nowait)
   {
      long javaLock = lock.isUnknown() ? LockType.LT_SHARE : lock.longValue();

      LockType lockType = fromLockNowait(javaLock, nowait);
      // toJavaType() returns null if predicate is unknown
      return lockType == null ?
            new logical(false) :
            _find(predicate.toJavaType(), FindMode.FIRST, lockType);
   }

   /**
    * Conversion of FIND-FIRST() method (KW_FIND_1ST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   predicate
    *          Predicate expression for find.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          Integer value of one of the following: NO-WAIT, 0, or the Unknown value (?).
    *
    * @return  <code>true</code> if FIND-FIRST() succeeds, else <code>false</code>.
    *
    * @throws  ErrorConditionException
    */
   public logical findFirst(character predicate, long lock, NumberType nowait)
   {
      LockType lockType = fromLockNowait(lock, nowait.longValue());
      // toJavaType() returns null if predicate is unknown
      return lockType == null ?
            new logical(false) :
            _find(predicate.toJavaType(), FindMode.FIRST, lockType);
   }

   /**
    * Conversion of FIND-FIRST() method (KW_FIND_1ST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   predicate
    *          Predicate expression for find.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          Integer value of one of the following: NO-WAIT, 0, or the Unknown value (?).
    *
    * @return  <code>true</code> if FIND-FIRST() succeeds, else <code>false</code>.
    *
    * @throws  ErrorConditionException
    */
   public logical findFirst(character predicate, NumberType lock, NumberType nowait)
   {
      long javaLock = lock.isUnknown() ? LockType.LT_SHARE : lock.longValue();

      LockType lockType = fromLockNowait(javaLock, nowait.longValue());
      // toJavaType() returns null if predicate is unknown
      return lockType == null ?
            new logical(false) :
            _find(predicate.toJavaType(), FindMode.FIRST, lockType);
   }

   /**
    * Conversion of FIND-LAST() method (KW_FIND_LST).
    *
    * Gets the last record.
    *
    * @return  <code>true</code> if FIND-LAST() succeeds, else <code>false</code>.
    */
   public logical findLast()
   {
      return _find(null, FindMode.LAST, LockType.SHARE);
   }

   /**
    * Conversion of FIND-LAST() method (KW_FIND_LST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   predicate
    *          Predicate expression for find.
    *
    * @return  <code>true</code> if FIND-LAST() succeeds, else <code>false</code>.
    */
   public logical findLast(character predicate)
   {
      // toJavaType() returns null if predicate is unknown
      return _find(predicate.toJavaType(), FindMode.LAST, LockType.SHARE);
   }

   /**
    * Conversion of FIND-LAST() method (KW_FIND_LST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   predicate
    *          Predicate expression for find.
    * @param   lockType
    *          Lock type
    *
    * @return  <code>true</code> if FIND-LAST() succeeds, else <code>false</code>.
    */
   public logical findLast(character predicate, LockType lockType)
   {
      // toJavaType() returns null if predicate is unknown
      return _find(predicate.toJavaType(), FindMode.LAST, lockType);
   }

   /**
    * Conversion of FIND-LAST() method (KW_FIND_LST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   predicate
    *          Predicate expression for find.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if FIND-LAST() succeeds, else <code>false</code>.
    */
   public logical findLast(character predicate, long lock)
   {
      // toJavaType() returns null if predicate is unknown
      return _find(predicate.toJavaType(), FindMode.LAST, LockType.fromInt((int) lock));
   }

   /**
    * Conversion of FIND-LAST() method (KW_FIND_LST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   predicate
    *          Predicate expression for find.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if FIND-LAST() succeeds, else <code>false</code>.
    */
   public logical findLast(character predicate, NumberType lock)
   {
      // toJavaType() returns null if predicate is unknown
      return _find(predicate.toJavaType(), FindMode.LAST, LockType.fromInt(lock.intValue()));
   }

   /**
    * Conversion of FIND-LAST() method (KW_FIND_LST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   predicate
    *          Predicate expression for find.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          Integer value of one of the following: NO-WAIT, 0, or the Unknown value (?).
    *
    * @return  <code>true</code> if FIND-LAST() succeeds, else <code>false</code>.
    *
    * @throws  ErrorConditionException
    */
   public logical findLast(character predicate, long lock, long nowait)
   {
      LockType lockType = fromLockNowait(lock, nowait);
      // toJavaType() returns null if predicate is unknown
      return lockType == null ?
            new logical(false) :
            _find(predicate.toJavaType(), FindMode.LAST, lockType);
   }

   /**
    * Conversion of FIND-LAST() method (KW_FIND_LST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   predicate
    *          Predicate expression for find.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          Integer value of one of the following: NO-WAIT, 0, or the Unknown value (?).
    *
    * @return  <code>true</code> if FIND-LAST() succeeds, else <code>false</code>.
    *
    * @throws  ErrorConditionException
    */
   public logical findLast(character predicate, NumberType lock, long nowait)
   {
      LockType lockType = fromLockNowait(lock.longValue(), nowait);
      // toJavaType() returns null if predicate is unknown
      return lockType == null ?
            new logical(false) :
            _find(predicate.toJavaType(), FindMode.LAST, lockType);
   }

   /**
    * Conversion of FIND-LAST() method (KW_FIND_LST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   predicate
    *          Predicate expression for find.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          Integer value of one of the following: NO-WAIT, 0, or the Unknown value (?).
    *
    * @return  <code>true</code> if FIND-LAST() succeeds, else <code>false</code>.
    *
    * @throws  ErrorConditionException
    */
   public logical findLast(character predicate, long lock, NumberType nowait)
   {
      LockType lockType = fromLockNowait(lock, nowait.longValue());
      // toJavaType() returns null if predicate is unknown
      return lockType == null ?
            new logical(false) :
            _find(predicate.toJavaType(), FindMode.LAST, lockType);
   }

   /**
    * Conversion of FIND-LAST() method (KW_FIND_LST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   predicate
    *          Predicate expression for find.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          Integer value of one of the following: NO-WAIT, 0, or the Unknown value (?).
    *
    * @return  <code>true</code> if FIND-LAST() succeeds, else <code>false</code>.
    *
    * @throws  ErrorConditionException
    */
   public logical findLast(character predicate, NumberType lock, NumberType nowait)
   {
      LockType lockType = fromLockNowait(lock.longValue(), nowait.longValue());
      // toJavaType() returns null if predicate is unknown
      return lockType == null ?
            new logical(false) :
            _find(predicate.toJavaType(), FindMode.LAST, lockType);
   }

   /**
    * Conversion of FIND-UNIQUE() method (KW_FIND_UNI).
    *
    * Gets a single record. This make sense only when there is only one record in the table.
    * Otherwise the error 3166 "More than one <tablename> records found by a unique FIND."
    *
    * @return  <code>true</code> if FIND-UNIQUE() succeeds, else <code>false</code>.
    */
   public logical findUnique()
   {
      return _find(null, FindMode.UNIQUE, LockType.SHARE);
   }

   /**
    * Conversion of FIND-UNIQUE() method (KW_FIND_UNI).
    *
    * Gets a single record. This method lets a user get a unique record that
    * satisfies the predicate expression.
    *
    * @param   predicate
    *          Predicate expression for find.
    *
    * @return  <code>true</code> if FIND-UNIQUE() succeeds, else <code>false</code>.
    */
   public logical findUnique(character predicate)
   {
      // toJavaType() returns null if predicate is unknown
      return _find(predicate.toJavaType(), FindMode.UNIQUE, LockType.SHARE);
   }

   /**
    * Conversion of FIND-UNIQUE() method (KW_FIND_UNI).
    *
    * Gets a single record. This method lets a user get a unique record that
    * satisfies the predicate expression.
    *
    * @param   predicate
    *          Predicate expression for find.
    * @param   lockType
    *          Lock type
    *
    * @return  <code>true</code> if FIND-UNIQUE() succeeds, else <code>false</code>.
    */
   public logical findUnique(character predicate, LockType lockType)
   {
      // toJavaType() returns null if predicate is unknown
      return _find(predicate.toJavaType(), FindMode.UNIQUE, lockType);
   }

   /**
    * Conversion of FIND-UNIQUE() method (KW_FIND_UNI).
    *
    * Gets a single record. This method lets a user get a unique record that
    * satisfies the predicate expression.
    *
    * @param   predicate
    *          Predicate expression for find.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if FIND-UNIQUE() succeeds, else <code>false</code>.
    */
   public logical findUnique(character predicate, long lock)
   {
      // toJavaType() returns null if predicate is unknown
      return _find(predicate.toJavaType(), FindMode.UNIQUE, LockType.fromInt((int) lock));
   }

   /**
    * Conversion of FIND-UNIQUE() method (KW_FIND_UNI).
    *
    * Gets a single record. This method lets a user get a unique record that
    * satisfies the predicate expression.
    *
    * @param   predicate
    *          Predicate expression for find.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if FIND-UNIQUE() succeeds, else <code>false</code>.
    */
   public logical findUnique(character predicate, NumberType lock)
   {
      // toJavaType() returns null if predicate is unknown
      return _find(predicate.toJavaType(), FindMode.UNIQUE, LockType.fromInt(lock.intValue()));
   }

   /**
    * Conversion of FIND-UNIQUE() method (KW_FIND_UNI).
    *
    * Gets a single record. This method lets a user get a unique record that
    * satisfies the predicate expression.
    *
    * @param   predicate
    *          Predicate expression for find.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          Integer value of one of the following: NO-WAIT, 0, or the Unknown value (?).
    *
    * @return  <code>true</code> if FIND-UNIQUE() succeeds, else <code>false</code>.
    *
    * @throws  ErrorConditionException
    */
   public logical findUnique(character predicate, long lock, long nowait)
   {
      LockType lockType = fromLockNowait(lock, nowait);
      // toJavaType() returns null if predicate is unknown
      return lockType == null ?
            new logical(false) :
            _find(predicate.toJavaType(), FindMode.UNIQUE, lockType);
   }

   /**
    * Conversion of FIND-UNIQUE() method (KW_FIND_UNI).
    *
    * Gets a single record. This method lets a user get a unique record that
    * satisfies the predicate expression.
    *
    * @param   predicate
    *          Predicate expression for find.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          Integer value of one of the following: NO-WAIT, 0, or the Unknown value (?).
    *
    * @return  <code>true</code> if FIND-UNIQUE() succeeds, else <code>false</code>.
    *
    * @throws  ErrorConditionException
    */
   public logical findUnique(character predicate, NumberType lock, long nowait)
   {
      LockType lockType = fromLockNowait(lock.longValue(), nowait);
      // toJavaType() returns null if predicate is unknown
      return lockType == null ?
            new logical(false) :
            _find(predicate.toJavaType(), FindMode.UNIQUE, lockType);
   }

   /**
    * Conversion of FIND-UNIQUE() method (KW_FIND_UNI).
    *
    * Gets a single record. This method lets a user get a unique record that
    * satisfies the predicate expression.
    *
    * @param   predicate
    *          Predicate expression for find.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          Integer value of one of the following: NO-WAIT, 0, or the Unknown value (?).
    *
    * @return  <code>true</code> if FIND-UNIQUE() succeeds,
    *          else <code>false</code>.
    *
    * @throws  ErrorConditionException
    */
   public logical findUnique(character predicate, long lock, NumberType nowait)
   {
      LockType lockType = fromLockNowait(lock, nowait.longValue());
      // toJavaType() returns null if predicate is unknown
      return lockType == null ?
            new logical(false) :
            _find(predicate.toJavaType(), FindMode.UNIQUE, lockType);
   }

   /**
    * Conversion of FIND-UNIQUE() method (KW_FIND_UNI).
    *
    * Gets a single record. This method lets a user get a unique record that
    * satisfies the predicate expression.
    *
    * @param   predicate
    *          Predicate expression for find.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          Integer value of one of the following: NO-WAIT, 0, or the Unknown value (?).
    *
    * @return  <code>true</code> if FIND-UNIQUE() succeeds,
    *          else <code>false</code>.
    *
    * @throws  ErrorConditionException
    */
   public logical findUnique(character predicate, NumberType lock, NumberType nowait)
   {
      LockType lockType = fromLockNowait(lock.longValue(), nowait.longValue());
      // toJavaType() returns null if predicate is unknown
      return lockType == null ?
            new logical(false) :
            _find(predicate.toJavaType(), FindMode.UNIQUE, lockType);
   }

   /**
    * Conversion of NUM-FIELDS attribute (KW_NUM_FLD).
    *
    * Getter of NUM-FIELDS attribute.
    *
    * @return  current value of NUM-FIELDS attribute.
    */
   public integer numFields()
   {
      return new integer(_numFields());
   }

   /**
    * Disable the triggers associated with the FIND event for the buffer's table.
    */
   public void disableDumpTriggers()
   {
      final Class<?> dmoBuf = buffer().getDMOBufInterface();
      if (Buffer.class.isAssignableFrom(dmoBuf))
      {
         DatabaseTriggerManager.disableDumpTriggers((Class<? extends Buffer>) dmoBuf);
      }
   }

   /**
    * Disable the triggers associated with the CREATE, WRITE, DELETE and ASSIGN events for the
    * buffer's table.
    *
    * @param allowReplication
    *        If <code>true</code> then only CREATE, DELETE, ASSIGN and WRITE triggers are
    *        disabled, and not REPLICATION-CREATE, REPLICATION-DELETE and REPLICATION-WRITE
    *        triggers.
    */
   public void disableLoadTriggers(boolean allowReplication)
   {
      final Class<?> dmoBuf = buffer().getDMOBufInterface();
      if (Buffer.class.isAssignableFrom(dmoBuf))
      {
         DatabaseTriggerManager.disableLoadTriggers((Class<? extends Buffer>) dmoBuf,
                                                    allowReplication);
      }
   }

   /**
    * Conversion of CURRENT-ITERATION attribute (KW_CUR_ITER).
    *
    * Getter of CURRENT-ITERATION attribute.
    *
    * @return current value of CURRENT-ITERATION attribute.
    */
   public integer currentIteration()
   {
      // TODO: implement, returning unknown values for the moment
      return new integer();
   }

   /**
    * Returns the state of the ERROR attribute
    *
    * @return    Current state of the ERROR attribute.
    */
   public logical isError()
   {
      return new logical(hasError);
   }

   /**
    * Setter for ERROR attribute.
    * The ERROR attribute is writable for a Buffer object. The attribute can be set
    * programatically in order to manage the application response to errors.
    * If set to FALSE, the application will ignore all errors that might be returned for a
    * statement using the NO-ERROR option.
    * However, manually setting this attribute won't raise the ERROR condition.
    *
    * @param    value
    *           The new value of ERROR attribute.
    */
   public void setError(boolean value)
   {
      final RecordBuffer buf = buffer();
      if (buf.getParentTable() == null)
      {
         // permanent table case
         ErrorManager.recordOrShowError(11917,
               "Cannot set ERROR if there is no BEFORE-TABLE for " + getTable().getValue(),
               false, false, true);
         ErrorManager.recordOrShowError(3131,
               "Unable to set attribute ERROR in widget  of type BUFFER",
               false, false, true);
         return;
      }

      // temp-tables case
      ErrorManager.recordOrShowError(11916,
               getTable().getValue() + " record is not available to set ERROR in",
               false, false, true);
      ErrorManager.recordOrShowError(3131,
               "Unable to set attribute ERROR in widget  of type BUFFER",
               false, false, true);

      // "Cannot set ERROR if there is no BEFORE-TABLE row for tt1. (11918)"

      // this.hasError = value;
   }

   /**
    * Setter for ERROR attribute.
    *
    * @param    value
    *           The new value of ERROR attribute.
    */
   public void setError(logical value)
   {
      setError(value.booleanValue());
   }

   /**
    * Returns the state of the DBNAME attribute
    *
    * @return    Current value of the DBNAME attribute.
    */
   public character getDbName()
   {
      final RecordBuffer buf = buffer();
      if (buf.getParentTable() != null)
      {
         // in the case of temp-tables (defined statically or dynamically)
         return new character("PROGRESS");
      }

      // permanent tables, it works as expected (the logical name of the database of the buffer)
      return new character(buf.getLogicalDatabase());
   }

   /**
    * Returns the state of the TABLE attribute
    *
    * @return    Current value of the TABLE attribute.
    */
   public character getTable()
   {
      // the buffer().getTable() because is NOT what we need here (the legacy name)
      return new character(TableMapper.getLegacyName(this));
      // alternately, we could return this.name;
   }

   /**
    * Conversion of TABLE-HANDLE attribute (KW_TAB_HAND).
    *
    * Gets the handle of a temp-table object, if any, associated with the buffer object.
    *
    * @return  the handle of a temp-table object.
    */
   public handle tableHandle()
   {
      return buffer().tableHandle();
   }

   /**
    * Conversion of DYNAMIC attribute. Returns <code>true</code> if the buffer is dynamic.
    *
    * @return <code>true</code> if the buffer is dynamic.
    */
   public boolean _dynamic()
   {
      return buffer().isDynamic();
   }

   /**
    * Conversion of DYNAMIC attribute. Returns <code>true</code> if the buffer is dynamic.
    *
    * @return <code>true</code> if the buffer is dynamic.
    */
   public logical dynamic()
   {
      return new logical(_dynamic());
   }

   /**
    * Get the value of the ADM-DATA attribute.
    * 
    * @return   See above.
    */
   @Override
   public character getADMData()
   {
      return new character(this.admData);
   }

   /**
    * Set the value of the ADM-DATA attribute.
    * 
    * @param    value
    *           The new value.
    */
   @Override
   public void setADMData(String value)
   {
      this.admData.assign(value);
   }

   /**
    * Set the value of the ADM-DATA attribute.
    * 
    * @param    value
    *           The new value.
    */
   @Override
   public void setADMData(character value)
   {
      this.admData.assign(value);
   }

   /**
    * Gets the the unique ID number associated to this object by the underlying system. Not the 
    * same as the handle value itself.
    * 
    * @return   An integer representing an unique id between all buffers ever created in the 
    *           current process.
    */
   @Override
   public integer getUniqueID()
   {
      return new integer(this.uniqueID);
   }

   /**
    * Returns the value of the the NAMESPACE-URI attribute.
    *
    * @return    See above.
    */
   @Override
   public character getNamespaceURI()
   {
      return new character(namespaceURI);
   }

   /**
    * Setter for NAMESPACE-URI attribute.
    *
    * @param    uri
    *           The new value of the NAMESPACE-URI attribute.
    */
   @Override
   public void setNamespaceURI(String uri)
   {
      namespaceURI = uri;
   }

   /**
    * Setter for NAMESPACE-URI attribute.
    *
    * @param    uri
    *           The new value of the NAMESPACE-URI attribute.
    */
   @Override
   public void setNamespaceURI(character uri)
   {
      setNamespaceURI(uri == null ? null : uri.getValue());
   }

   /**
    * Conversion of BUFFER-FIELD() method (KW_BUF_FLD).
    *
    * Gets the handle to a particular buffer of a query or ProDataSet object.
    *
    * @param   fieldNumber
    *          field number. An INTEGER expression representing the sequence
    *          number of the field in the buffer.
    *
    * @return  handle to a particular field in the buffer.
    */
   @Override
   public handle bufferField(NumberType fieldNumber)
   {
      if (fieldNumber == null || fieldNumber.isUnknown())
      {
         ErrorManager.recordOrShowError(7349,
             "Argument for BUFFER-FIELD method for buffer " + this.name.getValue() +
                   " could not be evaluated.",
             false, false, false);
         return new handle();
      }

      return bufferField(fieldNumber.intValue());
   }

   /**
    * Conversion of BUFFER-FIELD() method (KW_BUF_FLD).
    *
    * Gets the handle to a particular buffer of a query or ProDataSet object.
    *
    * @param   fieldName
    *          A CHARACTER string expression representing the name of the field in the buffer.
    *
    * @return  handle to a particular field in the buffer.
    */
   @Override
   public handle bufferField(character fieldName)
   {
      if (fieldName.isUnknown())
      {
         String errMsg1 =
            "Argument for BUFFER-FIELD method for buffer %s could not be evaluated.";
         ErrorManager.recordOrShowError(
            7349, String.format(errMsg1, name()), false, false, false);
         String errMsg2 =
            "Lead attributes in a chained-attribute expression (a:b:c) must be type " +
            "HANDLE or a user-defined type and valid (not UNKNOWN)";
         ErrorManager.recordOrShowError(10068, errMsg2, false, false, false);
         return new handle();
      }
      else
      {
         return bufferField(fieldName.getValue());
      }
   }

   /**
    * Set NAME attribute without any checks.
    *
    * @param name
    *        Name to set.
    */
   void doSetName(String name)
   {
      this.name.assign(name);
   }

   /**
    * Check if this resource supports the NEXT-SIBLING attribute.
    * 
    * @return   Always <code>false</code>
    */
   @Override
   protected boolean hasNextSibling()
   {
      return false;
   }

   /**
    * Delete the resource.
    * 
    * @return   <code>true</code> if the resource was deleted.
    */
   @Override
   protected boolean resourceDelete()
   {
      if (!implicitDeletion() && !_dynamic())
      {
         // static buffers can' be deleted, but no error message is shown
         return false;
      }

      buffer().finished();
      return true;
   }

   /**
    * Return LockType instance for the given internal lock and nowait with corresponding error
    * processing .
    * <p>
    *
    * @param   lock
    *          Integer value.
    * @param   nowait
    *          Integer value.
    *
    * @return  <code>LockType</code> instance.
    *
    * @throws  ErrorConditionException
    */
   private static LockType fromLockNowait(long lock, long nowait)
   throws ErrorConditionException
   {
      LockType lockType = null;
      try
      {
         lockType = LockType.fromLockNowait(lock, nowait);
      }
      catch (LockTypeException e)
      {
         String err;
         int num;
         if (e.isNowaitCorrect())
         {
            err = "FIND-BY-ROWID, FIND-FIRST etc., second argument must be " +
                  "NO-LOCK, SHARE-LOCK or EXCLUSIVE-LOCK";
            num = 7344;
         }
         else
         {
            err = "FIND method argument must be NO-WAIT, or 0 or ?";
            num = 7343;
         }
         ErrorManager.recordOrThrowError(num, err);
      }
      return lockType;
   }

   /**
    * Getter of NUM-FIELDS attribute.
    *
    * @return  current value of NUM-FIELDS attribute.
    */
   private int _numFields()
   {
      // we count the number of properties directly from PropertyHelper
      return PropertyHelper.gettersByProperty(buffer().getDMOInterface()).size();
   }

   /**
    * Returns the hibernate name of a legacy field name. This works for permanent tabeles, static
    * and dynamic temp-tables.
    *
    * @param   memberName
    *          The legacy field name.
    *
    * @return  The associated Hibernate field name.
    */
   private String _getHibernatePropertyName(String memberName)
   {
      String hibernateName = null;

      TempTable parentTempTable = buffer().getParentTable();
      if (parentTempTable != null && parentTempTable._dynamic())
      {
         // dynamic table
         hibernateName = DynamicTablesHelper.getHibernateFieldName(buffer().getDMOInterface(),
                                                                   memberName);
      }
      else
      {
         // static temporary or permanent table
         Database database = buffer().getDatabase();
         if (database.equals(DatabaseManager.TEMP_TABLE_DB))
         {
            hibernateName = TableMapper.getPropertyName(parentTempTable, memberName);
         }
         else
         {
            String schema = DatabaseManager.getSchema(database);
            String legacyTableName = TableMapper.getLegacyName(this);
            hibernateName = TableMapper.getPropertyName(schema, legacyTableName, memberName);
            // hibernateName should not be null as the memberName should have been already checked
            // at conversion time
         }
      }
      return hibernateName;
   }

   /**
    * Execute a dynamic FIRST, LAST or UNIQUE find query. All the public overloade methods
    * (findFirst, findLast and findUnique) will call this private implementation using the 
    * appropriate parameters. Internally this method uses the DynamicQueryHelper.parseFindQuery
    * static parser to return a FindQuery object using the passed predicate and locking mode.
    * After the FindQuery is obtained, the appropriate query is executed (FIRST, LAST or UNIQUE)
    * and the result is returned to caller method.
    *
    * @param   predicate
    *          Predicate expression for find that evaluates to the following syntax:
    *             [ WHERE [ logical-expression ] ] [ USE-INDEX index-name ]
    *          <ul>
    *             <li>Once evaluated, predicate-expression can contain only constants and
    *                   unabbreviated references to fields from the buffer.
    *             <li>The predicate-expression itself can be built using a concatenation of
    *                   character expressions.
    *          </ul>
    * @param   mode
    *          Search mode, one of FIRST, LAST or UNIQUE.
    * @param   lock
    *          Lock type. May additionally have no-wait flag set.
    *
    * @return  <code>true</code> if operation succeeds, else <code>false</code>.
    */
   private logical _find(String predicate, FindMode mode, LockType lock)
   {
      final String qName = "find_" + mode + "_" + lock;
      P2JQuery query = DynamicQueryHelper.parseFindQuery(this, predicate, lock, qName);
      if (query == null)
      {
         ErrorManager.recordOrShowError(10041,
            "Invalid WHERE clause in FIND method for buffer " + buffer().getTable(),
            false, false, false);
         return new logical(false);
      }

      boolean success = true;
      try
      {
         switch (mode)
         {
            case FIRST:
               query.first();
               break;

            case LAST:
               query.last();
               break;

            case UNIQUE:
               query.unique();
               break;
         }
         success = !ErrorManager.isErrorFlag();
      }
      catch (Exception e)
      {
         success = false;
      }

      // the query is short-lived, we don't need it after this point
      DynamicQueryHelper.unregisterQuery(query);

      return new logical(success);
   }

}
